"""
Flow run command — native cyclopts implementation.

Interact with flow runs.
"""

from __future__ import annotations

import logging
import os
import signal
import threading
import webbrowser
from types import FrameType
from typing import TYPE_CHECKING, Annotated, Optional
from uuid import UUID

import cyclopts
import httpx
import orjson
from rich.markup import escape
from rich.pretty import Pretty
from rich.table import Table
from starlette import status

import prefect.cli._app as _cli
from prefect.cli._utilities import (
    exit_with_error,
    exit_with_success,
    with_cli_exception_handling,
)
from prefect.client.orchestration import get_client
from prefect.client.schemas.filters import FlowFilter, FlowRunFilter, LogFilter
from prefect.client.schemas.objects import StateType
from prefect.client.schemas.responses import SetStateStatus
from prefect.client.schemas.sorting import FlowRunSort, LogSort
from prefect.exceptions import ObjectNotFound
from prefect.logging import get_logger
from prefect.runner import Runner
from prefect.states import State, exception_to_crashed_state
from prefect.types._datetime import human_friendly_diff
from prefect.utilities.asyncutils import run_sync_in_worker_thread
from prefect.utilities.urls import url_for

if TYPE_CHECKING:
    from prefect.client.orchestration import PrefectClient
    from prefect.client.schemas.objects import FlowRun

flow_run_app: cyclopts.App = cyclopts.App(
    name="flow-run",
    alias="flow-runs",
    help="Interact with flow runs.",
    version_flags=[],
    help_flags=["--help"],
)

LOGS_DEFAULT_PAGE_SIZE = 200
LOGS_WITH_LIMIT_FLAG_DEFAULT_NUM_LOGS = 20

logger: logging.Logger = get_logger(__name__)


async def _get_flow_run_by_id_or_name(
    client: PrefectClient,
    id_or_name: str,
) -> FlowRun:
    """Resolve a flow run identifier that could be either a UUID or a name."""
    from prefect.client.schemas.filters import FlowRunFilterName

    try:
        flow_run_id = UUID(id_or_name)
        try:
            return await client.read_flow_run(flow_run_id)
        except ObjectNotFound:
            exit_with_error(f"Flow run '{id_or_name}' not found!")
    except ValueError:
        pass

    flow_runs = await client.read_flow_runs(
        flow_run_filter=FlowRunFilter(name=FlowRunFilterName(any_=[id_or_name])),
        limit=100,
    )

    if not flow_runs:
        exit_with_error(f"Flow run '{id_or_name}' not found!")

    if len(flow_runs) == 1:
        return flow_runs[0]

    lines = [f"Multiple flow runs found with name '{id_or_name}':\n"]
    for fr in flow_runs:
        state_name = fr.state.name if fr.state else "unknown"
        timestamp = fr.start_time or fr.created
        timestamp_str = timestamp.strftime("%Y-%m-%d %H:%M:%S") if timestamp else "N/A"
        lines.append(f"  - {fr.id} ({state_name}, {timestamp_str})")

    lines.append("\nPlease retry using an explicit flow run ID.")
    exit_with_error("\n".join(lines))


@flow_run_app.command()
@with_cli_exception_handling
async def inspect(
    id: UUID,
    *,
    web: Annotated[
        bool,
        cyclopts.Parameter("--web", help="Open the flow run in a web browser."),
    ] = False,
    output: Annotated[
        Optional[str],
        cyclopts.Parameter(
            "--output",
            alias="-o",
            help="Specify an output format. Currently supports: json",
        ),
    ] = None,
):
    """View details about a flow run."""
    if output and output.lower() != "json":
        exit_with_error("Only 'json' output format is supported.")

    async with get_client() as client:
        try:
            flow_run = await client.read_flow_run(id)
        except httpx.HTTPStatusError as exc:
            if exc.response.status_code == status.HTTP_404_NOT_FOUND:
                exit_with_error(f"Flow run {id!r} not found!")
            else:
                raise

    if web:
        flow_run_url = url_for("flow-run", obj_id=id)
        if not flow_run_url:
            exit_with_error(
                "Failed to generate URL for flow run. Make sure PREFECT_UI_URL is configured."
            )

        await run_sync_in_worker_thread(webbrowser.open_new_tab, flow_run_url)
        exit_with_success(f"Opened flow run {id!r} in browser.")
    else:
        if output and output.lower() == "json":
            flow_run_json = flow_run.model_dump(mode="json")
            json_output = orjson.dumps(
                flow_run_json, option=orjson.OPT_INDENT_2
            ).decode()
            _cli.console.print(json_output, soft_wrap=True)
        else:
            _cli.console.print(Pretty(flow_run))


@flow_run_app.command()
@with_cli_exception_handling
async def ls(
    *,
    flow_name: Annotated[
        Optional[list[str]],
        cyclopts.Parameter("--flow-name", help="Name of the flow"),
    ] = None,
    limit: Annotated[
        int,
        cyclopts.Parameter("--limit", help="Maximum number of flow runs to list"),
    ] = 15,
    state: Annotated[
        Optional[list[str]],
        cyclopts.Parameter("--state", help="Name of the flow run's state"),
    ] = None,
    state_type: Annotated[
        Optional[list[str]],
        cyclopts.Parameter("--state-type", help="Type of the flow run's state"),
    ] = None,
    output: Annotated[
        Optional[str],
        cyclopts.Parameter(
            "--output",
            alias="-o",
            help="Specify an output format. Currently supports: json",
        ),
    ] = None,
):
    """View recent flow runs or flow runs for specific flows."""
    if output and output.lower() != "json":
        exit_with_error("Only 'json' output format is supported.")

    prefect_state_names = {
        "SCHEDULED": "Scheduled",
        "PENDING": "Pending",
        "RUNNING": "Running",
        "COMPLETED": "Completed",
        "FAILED": "Failed",
        "CANCELLED": "Cancelled",
        "CRASHED": "Crashed",
        "PAUSED": "Paused",
        "CANCELLING": "Cancelling",
        "SUSPENDED": "Suspended",
        "AWAITINGRETRY": "AwaitingRetry",
        "RETRYING": "Retrying",
        "LATE": "Late",
    }

    state_filter = {}
    formatted_states = []

    if state:
        for s in state:
            uppercased_state = s.upper()
            if uppercased_state in prefect_state_names:
                capitalized_state = prefect_state_names[uppercased_state]
                formatted_states.append(capitalized_state)
            else:
                formatted_states.append(s)
                logger.warning(
                    f"State name {repr(s)} is not one of the official Prefect state names."
                )

        state_filter["name"] = {"any_": formatted_states}

    if state_type:
        upper_cased_states = [s.upper() for s in state_type]
        if not all(s in StateType.__members__ for s in upper_cased_states):
            exit_with_error(
                f"Invalid state type. Options are {', '.join(StateType.__members__)}."
            )

        state_filter["type"] = {
            "any_": [StateType[s].value for s in upper_cased_states]
        }

    async with get_client() as client:
        flow_runs = await client.read_flow_runs(
            flow_filter=FlowFilter(name={"any_": flow_name}) if flow_name else None,
            flow_run_filter=FlowRunFilter(state=state_filter) if state_filter else None,
            limit=limit,
            sort=FlowRunSort.EXPECTED_START_TIME_DESC,
        )
        flows_by_id = {
            flow.id: flow
            for flow in await client.read_flows(
                flow_filter=FlowFilter(id={"any_": [run.flow_id for run in flow_runs]})
            )
        }

        if not flow_runs:
            if output and output.lower() == "json":
                _cli.console.print("[]")
                return
            exit_with_success("No flow runs found.")

    if output and output.lower() == "json":
        flow_runs_json = [flow_run.model_dump(mode="json") for flow_run in flow_runs]
        json_output = orjson.dumps(flow_runs_json, option=orjson.OPT_INDENT_2).decode()
        _cli.console.print(json_output, soft_wrap=True)
    else:
        table = Table(title="Flow Runs")
        table.add_column("ID", justify="right", style="cyan", no_wrap=True)
        table.add_column("Flow", style="blue", no_wrap=True)
        table.add_column("Name", style="green", no_wrap=True)
        table.add_column("State", no_wrap=True)
        table.add_column("When", style="bold", no_wrap=True)

        for flow_run in sorted(flow_runs, key=lambda d: d.created, reverse=True):
            flow = flows_by_id[flow_run.flow_id]
            timestamp = (
                flow_run.state.state_details.scheduled_time
                if flow_run.state.is_scheduled()
                else flow_run.state.timestamp
            )
            table.add_row(
                str(flow_run.id),
                str(flow.name),
                str(flow_run.name),
                str(flow_run.state.type.value),
                human_friendly_diff(timestamp),
            )

        _cli.console.print(table)


@flow_run_app.command()
@with_cli_exception_handling
async def delete(id: UUID):
    """Delete a flow run by ID."""
    from prefect.cli._prompts import confirm

    async with get_client() as client:
        try:
            if _cli.is_interactive() and not confirm(
                f"Are you sure you want to delete flow run with id {id!r}?",
                default=False,
            ):
                exit_with_error("Deletion aborted.")
            await client.delete_flow_run(id)
        except ObjectNotFound:
            exit_with_error(f"Flow run '{id}' not found!")

    exit_with_success(f"Successfully deleted flow run '{id}'.")


@flow_run_app.command()
@with_cli_exception_handling
async def cancel(id: UUID):
    """Cancel a flow run by ID."""
    async with get_client() as client:
        cancelling_state = State(type=StateType.CANCELLING)
        try:
            result = await client.set_flow_run_state(
                flow_run_id=id, state=cancelling_state
            )
        except ObjectNotFound:
            exit_with_error(f"Flow run '{id}' not found!")

    if result.status == SetStateStatus.ABORT:
        exit_with_error(
            f"Flow run '{id}' was unable to be cancelled. Reason:"
            f" '{result.details.reason}'"
        )

    exit_with_success(f"Flow run '{id}' was successfully scheduled for cancellation.")


@flow_run_app.command()
@with_cli_exception_handling
async def retry(
    id_or_name: str,
    *,
    entrypoint: Annotated[
        Optional[str],
        cyclopts.Parameter(
            "--entrypoint",
            alias="-e",
            help=(
                "The path to a file containing the flow to run, and the name of the flow "
                "function, in the format `path/to/file.py:flow_function_name`. "
                "Required if the flow run does not have an associated deployment."
            ),
        ),
    ] = None,
):
    """Retry a failed or completed flow run."""
    from prefect.flow_engine import run_flow
    from prefect.flows import InfrastructureBoundFlow, load_flow_from_entrypoint
    from prefect.states import Scheduled
    from prefect.utilities.callables import (
        get_call_parameters,
        parameters_to_args_kwargs,
    )

    terminal_states = {
        StateType.COMPLETED,
        StateType.FAILED,
        StateType.CANCELLED,
        StateType.CRASHED,
    }

    async with get_client() as client:
        flow_run = await _get_flow_run_by_id_or_name(client, id_or_name)
        flow_run_id = flow_run.id

        if flow_run.state is None or flow_run.state.type not in terminal_states:
            current_state = flow_run.state.type.value if flow_run.state else "unknown"
            exit_with_error(
                f"Flow run '{flow_run_id}' is in state '{current_state}' and cannot be retried. "
                f"Only flow runs in terminal states (COMPLETED, FAILED, CANCELLED, CRASHED) can be retried."
            )

        if flow_run.deployment_id:
            scheduled_state = Scheduled(message="Retried via CLI")
            try:
                result = await client.set_flow_run_state(
                    flow_run_id=flow_run_id, state=scheduled_state, force=True
                )
            except ObjectNotFound:
                exit_with_error(f"Flow run '{flow_run_id}' not found!")

            if result.status == SetStateStatus.ABORT:
                exit_with_error(
                    f"Flow run '{flow_run_id}' could not be retried. Reason: '{result.details.reason}'"
                )

            exit_with_success(
                f"Flow run '{flow_run_id}' has been scheduled for retry. "
                "A worker will pick it up shortly."
            )
        else:
            if not entrypoint:
                exit_with_error(
                    f"Flow run '{flow_run_id}' does not have an associated deployment. "
                    "Please provide an --entrypoint to the flow code.\n\n"
                    f"Example: prefect flow-run retry {flow_run_id} --entrypoint ./flows/my_flow.py:my_flow"
                )

            try:
                loaded_flow = load_flow_from_entrypoint(
                    entrypoint, use_placeholder_flow=False
                )
            except Exception as exc:
                exit_with_error(
                    f"Failed to load flow from entrypoint '{entrypoint}': {exc}"
                )

            if isinstance(loaded_flow, InfrastructureBoundFlow):
                _cli.console.print(
                    f"Retrying flow run '{flow_run_id}' on remote infrastructure "
                    f"(work pool: {loaded_flow.work_pool})..."
                )

                try:
                    await loaded_flow.retry(flow_run)
                except Exception as exc:
                    exit_with_error(f"Flow run failed: {exc}")

                flow_run = await client.read_flow_run(flow_run_id)
                final_state = flow_run.state.type.value if flow_run.state else "unknown"

                if flow_run.state and flow_run.state.is_completed():
                    exit_with_success(
                        f"Flow run '{flow_run_id}' completed successfully."
                    )
                else:
                    exit_with_error(
                        f"Flow run '{flow_run_id}' finished with state: {final_state}"
                    )
            else:
                scheduled_state = Scheduled(message="Retried via CLI (local execution)")
                try:
                    result = await client.set_flow_run_state(
                        flow_run_id=flow_run_id, state=scheduled_state, force=True
                    )
                except ObjectNotFound:
                    exit_with_error(f"Flow run '{flow_run_id}' not found!")

                if result.status == SetStateStatus.ABORT:
                    exit_with_error(
                        f"Flow run '{flow_run_id}' could not be retried. Reason: '{result.details.reason}'"
                    )

                _cli.console.print(f"Executing flow run '{flow_run_id}' locally...")

                flow_run = await client.read_flow_run(flow_run_id)

                try:
                    call_args, call_kwargs = parameters_to_args_kwargs(
                        loaded_flow.fn, flow_run.parameters if flow_run else {}
                    )
                    parameters = get_call_parameters(
                        loaded_flow.fn, call_args, call_kwargs
                    )
                except Exception as exc:
                    crashed = await exception_to_crashed_state(exc)
                    await client.set_flow_run_state(
                        flow_run_id=flow_run_id, state=crashed, force=True
                    )
                    exit_with_error(
                        "Failed to use parameters from previous attempt. Please ensure the flow signature has not changed since the last run."
                    )

                try:
                    run_flow(
                        flow=loaded_flow,
                        flow_run=flow_run,
                        return_type="state",
                        parameters=parameters,
                    )
                except Exception as exc:
                    exit_with_error(f"Flow run failed: {exc}")

                flow_run = await client.read_flow_run(flow_run_id)
                final_state = flow_run.state.type.value if flow_run.state else "unknown"

                if flow_run.state and flow_run.state.is_completed():
                    exit_with_success(
                        f"Flow run '{flow_run_id}' completed successfully."
                    )
                else:
                    exit_with_error(
                        f"Flow run '{flow_run_id}' finished with state: {final_state}"
                    )


@flow_run_app.command()
@with_cli_exception_handling
async def logs(
    id: UUID,
    *,
    head: Annotated[
        bool,
        cyclopts.Parameter(
            "--head",
            alias="-h",
            help=(
                f"Show the first {LOGS_WITH_LIMIT_FLAG_DEFAULT_NUM_LOGS} logs instead of"
                " all logs."
            ),
        ),
    ] = False,
    num_logs: Annotated[
        Optional[int],
        cyclopts.Parameter(
            "--num-logs",
            alias="-n",
            help=(
                "Number of logs to show when using the --head or --tail flag. If None,"
                f" defaults to {LOGS_WITH_LIMIT_FLAG_DEFAULT_NUM_LOGS}."
            ),
        ),
    ] = None,
    reverse: Annotated[
        bool,
        cyclopts.Parameter(
            "--reverse",
            alias="-r",
            help="Reverse the logs order to print the most recent logs first",
        ),
    ] = False,
    tail: Annotated[
        bool,
        cyclopts.Parameter(
            "--tail",
            alias="-t",
            help=(
                f"Show the last {LOGS_WITH_LIMIT_FLAG_DEFAULT_NUM_LOGS} logs instead of"
                " all logs."
            ),
        ),
    ] = False,
):
    """View logs for a flow run."""
    offset = 0
    more_logs = True
    num_logs_returned = 0

    if head and tail:
        exit_with_error("Please provide either a `head` or `tail` option but not both.")

    user_specified_num_logs = (
        num_logs or LOGS_WITH_LIMIT_FLAG_DEFAULT_NUM_LOGS
        if head or tail or num_logs
        else None
    )

    if tail:
        offset = max(0, user_specified_num_logs - LOGS_DEFAULT_PAGE_SIZE)

    log_filter = LogFilter(flow_run_id={"any_": [id]})

    async with get_client() as client:
        try:
            flow_run = await client.read_flow_run(id)
        except ObjectNotFound:
            exit_with_error(f"Flow run '{id!s}' not found!")

        while more_logs:
            num_logs_to_return_from_page = (
                LOGS_DEFAULT_PAGE_SIZE
                if user_specified_num_logs is None
                else min(
                    LOGS_DEFAULT_PAGE_SIZE, user_specified_num_logs - num_logs_returned
                )
            )

            page_logs = await client.read_logs(
                log_filter=log_filter,
                limit=num_logs_to_return_from_page,
                offset=offset,
                sort=(
                    LogSort.TIMESTAMP_DESC if reverse or tail else LogSort.TIMESTAMP_ASC
                ),
            )

            for log in reversed(page_logs) if tail and not reverse else page_logs:
                timestamp = f"{log.timestamp:%Y-%m-%d %H:%M:%S.%f}"[:-3]
                log_level = f"{logging.getLevelName(log.level):7s}"
                flow_run_info = f"Flow run {flow_run.name!r} - {escape(log.message)}"

                log_message = f"{timestamp} | {log_level} | {flow_run_info}"
                _cli.console.print(
                    log_message,
                    soft_wrap=True,
                )

            num_logs_returned += num_logs_to_return_from_page

            if tail:
                if offset != 0:
                    offset = (
                        0
                        if offset < LOGS_DEFAULT_PAGE_SIZE
                        else offset - LOGS_DEFAULT_PAGE_SIZE
                    )
                else:
                    more_logs = False
            else:
                if len(page_logs) == LOGS_DEFAULT_PAGE_SIZE:
                    offset += LOGS_DEFAULT_PAGE_SIZE
                else:
                    more_logs = False


@flow_run_app.command()
@with_cli_exception_handling
async def execute(
    id: Optional[UUID] = None,
):
    """Execute a flow run by ID."""
    if id is None:
        environ_flow_id = os.environ.get("PREFECT__FLOW_RUN_ID")
        if environ_flow_id:
            id = UUID(environ_flow_id)

    if id is None:
        exit_with_error("Could not determine the ID of the flow run to execute.")

    runner = Runner()

    def _handle_reschedule_sigterm(_signal: int, _frame: FrameType | None):
        logger.info("SIGTERM received, initiating graceful shutdown...")
        runner.reschedule_current_flow_runs()
        exit_with_success("Flow run successfully rescheduled.")

    on_sigterm = os.environ.get("PREFECT_FLOW_RUN_EXECUTE_SIGTERM_BEHAVIOR", "").lower()
    if (
        threading.current_thread() is threading.main_thread()
        and on_sigterm == "reschedule"
    ):
        signal.signal(signal.SIGTERM, _handle_reschedule_sigterm)

    await runner.execute_flow_run(id)



"""
The Telemetry service. Sends anonymous data to Prefect to help us improve.
"""

import logging
import os
import platform
from datetime import timedelta
from uuid import uuid4

import httpx
from docket import Perpetual

import prefect
from prefect.logging import get_logger
from prefect.server.database import PrefectDBInterface, provide_database_interface
from prefect.server.models import configuration
from prefect.server.schemas.core import Configuration
from prefect.server.services.perpetual_services import perpetual_service
from prefect.settings import PREFECT_DEBUG_MODE
from prefect.settings.context import get_current_settings
from prefect.types._datetime import now

logger: logging.Logger = get_logger(__name__)


async def _fetch_or_set_telemetry_session(
    db: PrefectDBInterface,
) -> tuple[str, str]:
    """
    Fetch or create a telemetry session in the configuration table.

    Returns:
        tuple of (session_start_timestamp, session_id)
    """
    async with db.session_context(begin_transaction=True) as session:
        telemetry_session = await configuration.read_configuration(
            session, "TELEMETRY_SESSION"
        )

        if telemetry_session is None:
            logger.debug("No telemetry session found, setting")
            session_id = str(uuid4())
            session_start_timestamp = now("UTC").isoformat()

            telemetry_session = Configuration(
                key="TELEMETRY_SESSION",
                value={
                    "session_id": session_id,
                    "session_start_timestamp": session_start_timestamp,
                },
            )

            await configuration.write_configuration(session, telemetry_session)
        else:
            logger.debug("Session information retrieved from database")
            session_id = telemetry_session.value["session_id"]
            session_start_timestamp = telemetry_session.value["session_start_timestamp"]

    logger.debug(f"Telemetry Session: {session_id}, {session_start_timestamp}")
    return (session_start_timestamp, session_id)


@perpetual_service(
    enabled_getter=lambda: get_current_settings().server.analytics_enabled,
    run_in_ephemeral=True,
    run_in_webserver=True,
)
async def send_telemetry_heartbeat(
    perpetual: Perpetual = Perpetual(automatic=True, every=timedelta(seconds=600)),
) -> None:
    """
    Sends anonymous telemetry data to Prefect to help us improve.

    It can be toggled off with the PREFECT_SERVER_ANALYTICS_ENABLED setting.
    """
    from prefect.client.constants import SERVER_API_VERSION

    db = provide_database_interface()
    session_start_timestamp, session_id = await _fetch_or_set_telemetry_session(db=db)
    telemetry_environment = os.environ.get(
        "PREFECT_API_TELEMETRY_ENVIRONMENT", "production"
    )

    heartbeat = {
        "source": "prefect_server",
        "type": "heartbeat",
        "payload": {
            "platform": platform.system(),
            "architecture": platform.machine(),
            "python_version": platform.python_version(),
            "python_implementation": platform.python_implementation(),
            "environment": telemetry_environment,
            "ephemeral_server": bool(os.getenv("PREFECT__SERVER_EPHEMERAL", False)),
            "api_version": SERVER_API_VERSION,
            "prefect_version": prefect.__version__,
            "session_id": session_id,
            "session_start_timestamp": session_start_timestamp,
        },
    }

    try:
        async with httpx.AsyncClient() as client:
            result = await client.post(
                "https://sens-o-matic.prefect.io/",
                json=heartbeat,
                headers={"x-prefect-event": "prefect_server"},
            )
        result.raise_for_status()
    except Exception as exc:
        logger.error(
            f"Failed to send telemetry: {exc}",
            exc_info=PREFECT_DEBUG_MODE.value(),
        )



# Import user-facing API
from typing import Any
from prefect.states import State
from prefect.logging import get_run_logger
from prefect.flows import FlowDecorator, flow, Flow, serve, aserve
from prefect.transactions import Transaction
from prefect.tasks import task, Task
from prefect.context import tags
from prefect.utilities.annotations import unmapped, allow_failure
from prefect._result_records import ResultRecordMetadata
from prefect.flow_runs import (
    apause_flow_run,
    aresume_flow_run,
    asuspend_flow_run,
    pause_flow_run,
    resume_flow_run,
    suspend_flow_run,
)
from prefect.client.orchestration import get_client
from prefect.client.cloud import get_cloud_client
import prefect.variables  # pyright: ignore[reportUnusedImport] # TODO: Does this need to be imported here?
import prefect.runtime  # pyright: ignore[reportUnusedImport] # TODO: Does this need to be imported here?

# Import modules that register types
import prefect.serializers  # pyright: ignore[reportUnusedImport]
import prefect.blocks.notifications  # pyright: ignore[reportUnusedImport]
import prefect.blocks.system  # pyright: ignore[reportUnusedImport]

# Initialize the process-wide profile and registry at import time
import prefect.context

# Perform any forward-ref updates needed for Pydantic models
import prefect.client.schemas

_types: dict[str, Any] = dict(
    Task=Task,
    Flow=Flow,
    ResultRecordMetadata=ResultRecordMetadata,
)
prefect.context.FlowRunContext.model_rebuild(_types_namespace=_types)
prefect.context.TaskRunContext.model_rebuild(_types_namespace=_types)
prefect.client.schemas.State.model_rebuild(_types_namespace=_types)
prefect.client.schemas.StateCreate.model_rebuild(_types_namespace=_types)
prefect.client.schemas.OrchestrationResult.model_rebuild(_types_namespace=_types)
Transaction.model_rebuild()

# Configure logging
import prefect.logging.configuration

prefect.logging.configuration.setup_logging()
prefect.logging.get_logger("profiles").debug(
    f"Using profile {prefect.context.get_settings_context().profile.name!r}"
)


from prefect._internal.compatibility.deprecated import (
    inject_renamed_module_alias_finder,
)

inject_renamed_module_alias_finder()

flow: FlowDecorator


# Declare API for type-checkers
__all__ = [
    "allow_failure",
    "apause_flow_run",
    "aresume_flow_run",
    "aserve",
    "asuspend_flow_run",
    "flow",
    "Flow",
    "get_client",
    "get_cloud_client",
    "get_run_logger",
    "pause_flow_run",
    "resume_flow_run",
    "serve",
    "State",
    "suspend_flow_run",
    "tags",
    "task",
    "Task",
    "Transaction",
    "unmapped",
]



"""
Config command — native cyclopts implementation.

Manages Prefect settings and profiles.
"""

import os
from pathlib import Path
from typing import Annotated, Any, Literal, Union, cast

import cyclopts
from dotenv import dotenv_values

import prefect.cli._app as _cli
import prefect.context
import prefect.settings
from prefect._internal.compatibility.backports import tomllib
from prefect.cli._utilities import exit_with_error, exit_with_success
from prefect.exceptions import ProfileSettingsValidationError
from prefect.settings import Settings
from prefect.settings.legacy import (
    Setting,
    _get_settings_fields,
    _get_valid_setting_names,
)
from prefect.utilities.annotations import NotSet
from prefect.utilities.collections import listrepr

config_app: cyclopts.App = cyclopts.App(
    name="config", help="View and set Prefect settings."
)


@config_app.command(name="set")
def set_(
    settings: Annotated[
        list[str], cyclopts.Parameter(help="Settings in VAR=VAL format")
    ],
) -> None:
    """
    Change the value for a setting by setting the value in the current profile.
    """
    valid_setting_names = _get_valid_setting_names(Settings)
    parsed_settings: dict[Union[str, Setting], Any] = {}

    for item in settings:
        try:
            setting, value = item.split("=", maxsplit=1)
        except ValueError:
            exit_with_error(
                f"Failed to parse argument {item!r}. Use the format 'VAR=VAL'."
            )

        if setting not in valid_setting_names:
            exit_with_error(f"Unknown setting name {setting!r}.")

        if setting in {"PREFECT_HOME", "PREFECT_PROFILES_PATH"}:
            exit_with_error(
                f"Setting {setting!r} cannot be changed with this command. "
                "Use an environment variable instead."
            )

        parsed_settings[setting] = value

    try:
        new_profile = prefect.settings.update_current_profile(parsed_settings)
    except ProfileSettingsValidationError as exc:
        help_message = ""
        for setting, problem in exc.errors:
            for error in problem.errors():
                help_message += (
                    f"[bold red]Validation error(s) for setting[/bold red] "
                    f"[blue]{setting.name}[/blue]\n\n - {error['msg']}\n\n"
                )
        exit_with_error(help_message)

    for setting, value in parsed_settings.items():
        _cli.console.print(f"Set {setting!r} to {value!r}.")
        if setting in os.environ:
            _cli.console.print(
                f"[yellow]{setting} is also set by an environment variable which will "
                f"override your config value. Run `unset {setting}` to clear it."
            )

    exit_with_success(f"Updated profile {new_profile.name!r}.")


@config_app.command()
def validate() -> None:
    """
    Read and validate the current profile.

    Deprecated settings will be automatically converted to new names.
    """
    profiles = prefect.settings.load_profiles()
    profile = profiles[prefect.context.get_settings_context().profile.name]

    profile.validate_settings()

    prefect.settings.save_profiles(profiles)
    exit_with_success("Configuration valid!")


@config_app.command()
def unset(
    setting_names: Annotated[
        list[str], cyclopts.Parameter(help="Setting names to unset")
    ],
    *,
    yes: Annotated[
        bool, cyclopts.Parameter("--yes", alias="-y", help="Skip confirmation")
    ] = False,
) -> None:
    """
    Restore the default value for a setting.

    Removes the setting from the current profile.
    """
    valid_setting_names = _get_valid_setting_names(Settings)
    settings_context = prefect.context.get_settings_context()
    profiles = prefect.settings.load_profiles()
    profile = profiles[settings_context.profile.name]
    parsed: set[Setting] = set()

    for setting_name in setting_names:
        if setting_name not in valid_setting_names:
            exit_with_error(f"Unknown setting name {setting_name!r}.")
        parsed.add(_get_settings_fields(Settings)[setting_name])

    for setting in parsed:
        if setting not in profile.settings:
            exit_with_error(f"{setting.name!r} is not set in profile {profile.name!r}.")

    if not yes and _cli.is_interactive():
        from rich.prompt import Confirm

        if not Confirm.ask(
            f"Are you sure you want to unset the following setting(s): "
            f"{listrepr(setting_names)}?",
            console=_cli.console,
        ):
            exit_with_error("Unset aborted.")

    profiles.update_profile(
        name=profile.name, settings={setting_name: None for setting_name in parsed}
    )

    for setting_name in setting_names:
        _cli.console.print(f"Unset {setting_name!r}.")
        if setting_name in os.environ:
            _cli.console.print(
                f"[yellow]{setting_name!r} is also set by an environment variable. "
                f"Use `unset {setting_name}` to clear it."
            )

    prefect.settings.save_profiles(profiles)
    exit_with_success(f"Updated profile {profile.name!r}.")


@config_app.command()
def view(
    *,
    show_defaults: Annotated[
        bool,
        cyclopts.Parameter(
            "--show-defaults",
            negative="--hide-defaults",
            help="Show default values",
        ),
    ] = False,
    show_sources: Annotated[
        bool,
        cyclopts.Parameter(
            "--show-sources",
            negative="--hide-sources",
            help="Show value sources",
        ),
    ] = True,
    show_secrets: Annotated[
        bool,
        cyclopts.Parameter(
            "--show-secrets",
            negative="--hide-secrets",
            help="Show secret values",
        ),
    ] = False,
) -> None:
    """
    Display the current settings.
    """
    valid_setting_names = _get_valid_setting_names(Settings)

    if show_secrets:
        dump_context = dict(include_secrets=True)
    else:
        dump_context = {}

    context = prefect.context.get_settings_context()
    current_profile_settings = context.profile.settings

    if ui_url := prefect.settings.PREFECT_UI_URL.value():
        _cli.console.print(
            f"\N{ROCKET} you are connected to:\n[green]{ui_url}[/green]",
            soft_wrap=True,
        )

    _cli.console.print(
        f"[bold blue]PREFECT_PROFILE={context.profile.name!r}[/bold blue]"
    )

    settings_output: list[str] = []
    processed_settings: set[str] = set()

    def _process_setting(
        setting: Setting,
        value: str,
        source: Literal[
            "env",
            "profile",
            "defaults",
            ".env file",
            "prefect.toml",
            "pyproject.toml",
        ],
    ):
        display_value = "********" if setting.is_secret and not show_secrets else value
        source_blurb = f" (from {source})" if show_sources else ""
        settings_output.append(f"{setting.name}='{display_value}'{source_blurb}")
        processed_settings.add(setting.name)

    def _collect_defaults(default_values: dict[str, Any], current_path: list[str]):
        for key, value in default_values.items():
            if isinstance(value, dict):
                _collect_defaults(cast(dict[str, Any], value), current_path + [key])
            else:
                setting = _get_settings_fields(Settings).get(
                    ".".join(current_path + [key])
                )
                if setting is None or setting.name in processed_settings:
                    continue
                _process_setting(setting, value, "defaults")

    def _process_toml_settings(
        settings: dict[str, Any],
        base_path: list[str],
        source: Literal["prefect.toml", "pyproject.toml"],
    ):
        for key, value in settings.items():
            if isinstance(value, dict):
                _process_toml_settings(
                    cast(dict[str, Any], value), base_path + [key], source
                )
            else:
                setting = _get_settings_fields(Settings).get(
                    ".".join(base_path + [key]), NotSet
                )
                if setting is NotSet:
                    continue
                elif (
                    isinstance(setting, Setting) and setting.name in processed_settings
                ):
                    continue
                elif isinstance(setting, Setting):
                    _process_setting(setting, value, source)

    # Environment variables
    for setting_name in valid_setting_names:
        setting = _get_settings_fields(Settings)[setting_name]
        if setting.name in processed_settings:
            continue
        if (env_value := os.getenv(setting.name)) is None:
            continue
        _process_setting(setting, env_value, "env")

    # .env file
    for key, value in dotenv_values(".env").items():
        if key in valid_setting_names:
            setting = _get_settings_fields(Settings)[key]
            if setting.name in processed_settings or value is None:
                continue
            _process_setting(setting, value, ".env file")

    # prefect.toml
    if Path("prefect.toml").exists():
        toml_settings = tomllib.loads(Path("prefect.toml").read_text(encoding="utf-8"))
        _process_toml_settings(toml_settings, base_path=[], source="prefect.toml")

    # pyproject.toml
    if Path("pyproject.toml").exists():
        pyproject_settings = tomllib.loads(
            Path("pyproject.toml").read_text(encoding="utf-8")
        )
        pyproject_settings = pyproject_settings.get("tool", {}).get("prefect", {})
        _process_toml_settings(
            pyproject_settings, base_path=[], source="pyproject.toml"
        )

    # Profile settings
    for setting, value in current_profile_settings.items():
        if setting.name not in processed_settings:
            _process_setting(setting, value, "profile")

    if show_defaults:
        _collect_defaults(
            Settings().model_dump(context=dump_context),
            current_path=[],
        )

    _cli.console.print("\n".join(sorted(settings_output)), soft_wrap=True)



"""
Utilities for injecting FastAPI dependencies.
"""

from __future__ import annotations

import logging
import re
from base64 import b64decode
from typing import Annotated, Any, Optional
from uuid import UUID

from docket import Docket as Docket_
from fastapi import Body, Depends, Header, HTTPException
from packaging.version import Version
from starlette.requests import Request

from prefect._internal.compatibility.starlette import status
from prefect.server import schemas
from prefect.settings import PREFECT_API_DEFAULT_LIMIT


def provide_request_api_version(
    x_prefect_api_version: str = Header(None),
) -> Version | None:
    if not x_prefect_api_version:
        return

    # parse version
    try:
        _, _, _ = [int(v) for v in x_prefect_api_version.split(".")]
    except ValueError:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=(
                "Invalid X-PREFECT-API-VERSION header format.Expected header in format"
                f" 'x.y.z' but received {x_prefect_api_version}"
            ),
        )
    return Version(x_prefect_api_version)


class EnforceMinimumAPIVersion:
    """
    FastAPI Dependency used to check compatibility between the version of the api
    and a given request.

    Looks for the header 'X-PREFECT-API-VERSION' in the request and compares it
    to the api's version. Rejects requests that are lower than the minimum version.
    """

    def __init__(self, minimum_api_version: str, logger: logging.Logger):
        self.minimum_api_version = minimum_api_version
        versions = [int(v) for v in minimum_api_version.split(".")]
        self.api_major: int = versions[0]
        self.api_minor: int = versions[1]
        self.api_patch: int = versions[2]
        self.logger = logger

    async def __call__(
        self,
        x_prefect_api_version: str = Header(None),
    ) -> None:
        request_version = x_prefect_api_version

        # if no version header, assume latest and continue
        if not request_version:
            return

        # parse version
        try:
            major, minor, patch = [int(v) for v in request_version.split(".")]
        except ValueError:
            await self._notify_of_invalid_value(request_version)
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail=(
                    "Invalid X-PREFECT-API-VERSION header format."
                    f"Expected header in format 'x.y.z' but received {request_version}"
                ),
            )

        if (major, minor, patch) < (self.api_major, self.api_minor, self.api_patch):
            await self._notify_of_outdated_version(request_version)
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail=(
                    f"The request specified API version {request_version} but this "
                    f"server requires version {self.minimum_api_version} or higher."
                ),
            )

    async def _notify_of_invalid_value(self, request_version: str):
        self.logger.error(
            f"Invalid X-PREFECT-API-VERSION header format: '{request_version}'"
        )

    async def _notify_of_outdated_version(self, request_version: str):
        self.logger.error(
            f"X-PREFECT-API-VERSION header specifies version '{request_version}' "
            f"but minimum allowed version is '{self.minimum_api_version}'"
        )


def LimitBody() -> Any:
    """
    A `fastapi.Depends` factory for pulling a `limit: int` parameter from the
    request body while determining the default from the current settings.
    """

    def get_limit(
        limit: int = Body(
            None,
            description="Defaults to PREFECT_API_DEFAULT_LIMIT if not provided.",
        ),
    ):
        default_limit = PREFECT_API_DEFAULT_LIMIT.value()
        limit = limit if limit is not None else default_limit
        if not limit >= 0:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="Invalid limit: must be greater than or equal to 0.",
            )
        if limit > default_limit:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail=f"Invalid limit: must be less than or equal to {default_limit}.",
            )
        return limit

    return Depends(get_limit)


def get_created_by(
    prefect_automation_id: Optional[UUID] = Header(None, include_in_schema=False),
    prefect_automation_name: Optional[str] = Header(None, include_in_schema=False),
) -> Optional[schemas.core.CreatedBy]:
    """A dependency that returns the provenance information to use when creating objects
    during this API call."""
    if prefect_automation_id and prefect_automation_name:
        try:
            display_value = b64decode(prefect_automation_name.encode()).decode()
        except Exception:
            display_value = None

        if display_value:
            return schemas.core.CreatedBy(
                id=prefect_automation_id,
                type="AUTOMATION",
                display_value=display_value,
            )

    return None


def get_updated_by(
    prefect_automation_id: Optional[UUID] = Header(None, include_in_schema=False),
    prefect_automation_name: Optional[str] = Header(None, include_in_schema=False),
) -> Optional[schemas.core.UpdatedBy]:
    """A dependency that returns the provenance information to use when updating objects
    during this API call."""
    if prefect_automation_id and prefect_automation_name:
        return schemas.core.UpdatedBy(
            id=prefect_automation_id,
            type="AUTOMATION",
            display_value=prefect_automation_name,
        )

    return None


def is_ephemeral_request(request: Request) -> bool:
    """
    A dependency that returns whether the request is to an ephemeral server.
    """
    return "ephemeral-prefect" in str(request.base_url)


PREFECT_CLIENT_USER_AGENT_PATTERN = re.compile(
    r"^prefect/(\d+\.\d+\.\d+(?:[a-z.+0-9]+)?) \(API \S+\)$"
)


def get_prefect_client_version(
    user_agent: Annotated[Optional[str], Header(include_in_schema=False)] = None,
) -> Optional[str]:
    """
    Attempts to parse out the Prefect client version from the User-Agent header.

    The Prefect client sets the User-Agent header like so:
      f"prefect/{prefect.__version__} (API {constants.SERVER_API_VERSION})"
    """
    if not user_agent:
        return None

    if client_version := PREFECT_CLIENT_USER_AGENT_PATTERN.match(user_agent):
        return client_version.group(1)
    return None


def docket(request: Request) -> Docket_:
    return request.app.state.docket


Docket = Annotated[Docket_, Depends(docket)]



"""
First-run telemetry notice for SDK telemetry.

Displays a notice to users the first time telemetry is enabled,
but only in interactive terminal sessions.
"""

import sys
from pathlib import Path

from prefect._internal.analytics.emit import _is_interactive_terminal
from prefect.settings import get_current_settings

NOTICE_TEXT = """
Prefect collects anonymous usage data to improve the product.
To opt out: set PREFECT_SERVER_ANALYTICS_ENABLED=false on the server, or DO_NOT_TRACK=1 in the client.
Learn more: https://docs.prefect.io/concepts/telemetry
"""


def _get_notice_marker_path() -> Path:
    """Get the path to the notice marker file."""
    settings = get_current_settings()
    return settings.home / ".sdk_telemetry" / "notice_shown"


def _has_shown_notice() -> bool:
    """Check if the notice has been shown before."""
    return _get_notice_marker_path().exists()


def _mark_notice_shown() -> None:
    """Mark that the notice has been shown."""
    marker_path = _get_notice_marker_path()
    try:
        marker_path.parent.mkdir(parents=True, exist_ok=True)
        marker_path.touch()
    except Exception:
        pass


def maybe_show_telemetry_notice() -> None:
    """
    Show the telemetry notice if appropriate.

    The notice is shown only:
    - The first time telemetry is enabled
    - In an interactive terminal (TTY)
    """
    # Only show in interactive terminals
    if not _is_interactive_terminal():
        return

    # Only show once
    if _has_shown_notice():
        return

    # Show notice and mark as shown
    print(NOTICE_TEXT, file=sys.stderr)
    _mark_notice_shown()



from __future__ import annotations

import abc
import asyncio
import concurrent.futures
import multiprocessing
import os
import sys
import threading
import uuid
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from contextvars import copy_context
from functools import partial
from types import CoroutineType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterable,
    Protocol,
    overload,
)

from typing_extensions import ParamSpec, Self, TypeVar

from prefect._internal.uuid7 import uuid7
from prefect.client.schemas.objects import RunInput
from prefect.events.schemas.events import Event
from prefect.events.worker import EventsWorker, ProcessPoolForwardingEventsClient
from prefect.exceptions import MappingLengthMismatch, MappingMissingIterable
from prefect.futures import (
    PrefectConcurrentFuture,
    PrefectDistributedFuture,
    PrefectFuture,
    PrefectFutureList,
    wait,
)
from prefect.logging.handlers import APILogWorker, set_api_log_sink
from prefect.logging.loggers import get_logger, get_run_logger
from prefect.settings.context import get_current_settings
from prefect.utilities.annotations import allow_failure, opaque, quote, unmapped
from prefect.utilities.callables import (
    cloudpickle_wrapped_call,
    collapse_variadic_parameters,
    explode_variadic_parameter,
    get_parameter_defaults,
)
from prefect.utilities.collections import isiterable

if TYPE_CHECKING:
    import logging

    from prefect.tasks import Task

P = ParamSpec("P")
T = TypeVar("T")
R = TypeVar("R")
F = TypeVar("F", bound=PrefectFuture[Any], default=PrefectConcurrentFuture[Any])


class TaskRunner(abc.ABC, Generic[F]):
    """
    Abstract base class for task runners.

    A task runner is responsible for submitting tasks to the task run engine running
    in an execution environment. Submitted tasks are non-blocking and return a future
    object that can be used to wait for the task to complete and retrieve the result.

    Task runners are context managers and should be used in a `with` block to ensure
    proper cleanup of resources.
    """

    def __init__(self):
        self.logger: "logging.Logger" = get_logger(f"task_runner.{self.name}")
        self._started = False

    @property
    def name(self) -> str:
        """The name of this task runner"""
        return type(self).__name__.lower().replace("taskrunner", "")

    @abc.abstractmethod
    def duplicate(self) -> Self:
        """Return a new instance of this task runner with the same configuration."""
        ...

    @overload
    @abc.abstractmethod
    def submit(
        self,
        task: "Task[P, CoroutineType[Any, Any, R]]",
        parameters: dict[str, Any],
        wait_for: Iterable[PrefectFuture[Any]] | None = None,
        dependencies: dict[str, set[RunInput]] | None = None,
    ) -> F: ...

    @overload
    @abc.abstractmethod
    def submit(
        self,
        task: "Task[Any, R]",
        parameters: dict[str, Any],
        wait_for: Iterable[PrefectFuture[Any]] | None = None,
        dependencies: dict[str, set[RunInput]] | None = None,
    ) -> F: ...

    @abc.abstractmethod
    def submit(
        self,
        task: "Task[P, R | CoroutineType[Any, Any, R]]",
        parameters: dict[str, Any],
        wait_for: Iterable[PrefectFuture[Any]] | None = None,
        dependencies: dict[str, set[RunInput]] | None = None,
    ) -> F: ...

    def map(
        self,
        task: "Task[P, R | CoroutineType[Any, Any, R]]",
        parameters: dict[str, Any | unmapped[Any] | allow_failure[Any]],
        wait_for: Iterable[PrefectFuture[R]] | None = None,
    ) -> PrefectFutureList[F]:
        """
        Submit multiple tasks to the task run engine.

        Args:
            task: The task to submit.
            parameters: The parameters to use when running the task.
            wait_for: A list of futures that the task depends on.

        Returns:
            An iterable of future objects that can be used to wait for the tasks to
            complete and retrieve the results.
        """
        if not self._started:
            raise RuntimeError(
                "The task runner must be started before submitting work."
            )

        from prefect.utilities.engine import (
            collect_task_run_inputs_sync,
            resolve_inputs_sync,
        )

        # We need to resolve some futures to map over their data, collect the upstream
        # links beforehand to retain relationship tracking.
        task_inputs = {
            k: collect_task_run_inputs_sync(v, max_depth=0)
            for k, v in parameters.items()
        }

        # Resolve the top-level parameters in order to get mappable data of a known length.
        # Nested parameters will be resolved in each mapped child where their relationships
        # will also be tracked.
        parameters = resolve_inputs_sync(parameters, max_depth=0)

        # Ensure that any parameters in kwargs are expanded before this check
        parameters = explode_variadic_parameter(task.fn, parameters)

        iterable_parameters: dict[str, Any] = {}
        static_parameters: dict[str, Any] = {}
        annotated_parameters: dict[str, Any] = {}
        for key, val in parameters.items():
            if isinstance(val, (allow_failure, opaque, quote)):
                # Unwrap annotated parameters to determine if they are iterable
                annotated_parameters[key] = val
                val = val.unwrap()

            if isinstance(val, unmapped):
                static_parameters[key] = val.value
            elif isiterable(val):
                iterable_parameters[key] = list(val)
            else:
                static_parameters[key] = val

        if not len(iterable_parameters):
            raise MappingMissingIterable(
                "No iterable parameters were received. Parameters for map must "
                f"include at least one iterable. Parameters: {parameters}"
            )

        iterable_parameter_lengths = {
            key: len(val) for key, val in iterable_parameters.items()
        }
        lengths = set(iterable_parameter_lengths.values())
        if len(lengths) > 1:
            raise MappingLengthMismatch(
                "Received iterable parameters with different lengths. Parameters for map"
                f" must all be the same length. Got lengths: {iterable_parameter_lengths}"
            )

        map_length = list(lengths)[0]

        futures: list[PrefectFuture[Any]] = []
        for i in range(map_length):
            call_parameters: dict[str, Any] = {
                key: value[i] for key, value in iterable_parameters.items()
            }
            call_parameters.update(
                {key: value for key, value in static_parameters.items()}
            )

            # Add default values for parameters; these are skipped earlier since they should
            # not be mapped over
            for key, value in get_parameter_defaults(task.fn).items():
                call_parameters.setdefault(key, value)

            # Re-apply annotations to each key again
            for key, annotation in annotated_parameters.items():
                call_parameters[key] = annotation.rewrap(call_parameters[key])

            # Collapse any previously exploded kwargs
            call_parameters = collapse_variadic_parameters(task.fn, call_parameters)

            futures.append(
                self.submit(
                    task=task,
                    parameters=call_parameters,
                    wait_for=wait_for,
                    dependencies=task_inputs,
                )
            )

        return PrefectFutureList(futures)

    def __enter__(self) -> Self:
        if self._started:
            raise RuntimeError("This task runner is already started")

        self.logger.debug("Starting task runner")
        self._started = True
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        self.logger.debug("Stopping task runner")
        self._started = False


class ThreadPoolTaskRunner(TaskRunner[PrefectConcurrentFuture[R]]):
    """
    A task runner that executes tasks in a separate thread pool.

    Attributes:
        max_workers: The maximum number of threads to use for executing tasks.
            Defaults to `PREFECT_TASK_RUNNER_THREAD_POOL_MAX_WORKERS` or `sys.maxsize`.

    Note:
        This runner uses `contextvars.copy_context()` for thread-safe context propagation.
        However, because contextvars are thread-local, frequent task submissions
        that modify context (e.g., using `prefect.tags` in a loop) can lead to
        new thread creation per task. This may cause an increase in threads and
        file descriptors, potentially hitting OS limits (`OSError: Too many open files`).
        If this occurs, consider minimizing context changes within looped tasks or
        adjusting system limits for open file descriptors.

    Examples:
        Use a thread pool task runner with a flow:

        ```python
        from prefect import flow, task
        from prefect.task_runners import ThreadPoolTaskRunner

        @task
        def some_io_bound_task(x: int) -> int:
            # making a query to a database, reading a file, etc.
            return x * 2

        @flow(task_runner=ThreadPoolTaskRunner(max_workers=3)) # use at most 3 threads at a time
        def my_io_bound_flow():
            futures = []
            for i in range(10):
                future = some_io_bound_task.submit(i * 100)
                futures.append(future)

            return [future.result() for future in futures]
        ```

        Use a thread pool task runner as a context manager:

        ```python
        from prefect.task_runners import ThreadPoolTaskRunner

        @task
        def some_io_bound_task(x: int) -> int:
            # making a query to a database, reading a file, etc.
            return x * 2

        # Use the runner directly
        with ThreadPoolTaskRunner(max_workers=2) as runner:
            future1 = runner.submit(some_io_bound_task, {"x": 1})
            future2 = runner.submit(some_io_bound_task, {"x": 2})

            result1 = future1.result()  # 2
            result2 = future2.result()  # 4
        ```

        Configure max workers via settings:

        ```python
        # Set via environment variable
        # export PREFECT_TASK_RUNNER_THREAD_POOL_MAX_WORKERS=8

        from prefect import flow
        from prefect.task_runners import ThreadPoolTaskRunner

        @flow(task_runner=ThreadPoolTaskRunner())  # Uses 8 workers from setting
        def my_flow():
            ...
        ```

    """

    def __init__(self, max_workers: int | None = None):
        super().__init__()
        current_settings = get_current_settings()
        self._executor: ThreadPoolExecutor | None = None
        self._max_workers = (
            (current_settings.tasks.runner.thread_pool_max_workers or sys.maxsize)
            if max_workers is None
            else max_workers
        )
        self._cancel_events: dict[uuid.UUID, threading.Event] = {}

    def duplicate(self) -> "ThreadPoolTaskRunner[R]":
        return type(self)(max_workers=self._max_workers)

    @overload
    def submit(
        self,
        task: "Task[P, CoroutineType[Any, Any, R]]",
        parameters: dict[str, Any],
        wait_for: Iterable[PrefectFuture[Any]] | None = None,
        dependencies: dict[str, set[RunInput]] | None = None,
    ) -> PrefectConcurrentFuture[R]: ...

    @overload
    def submit(
        self,
        task: "Task[Any, R]",
        parameters: dict[str, Any],
        wait_for: Iterable[PrefectFuture[Any]] | None = None,
        dependencies: dict[str, set[RunInput]] | None = None,
    ) -> PrefectConcurrentFuture[R]: ...

    def submit(
        self,
        task: "Task[P, R | CoroutineType[Any, Any, R]]",
        parameters: dict[str, Any],
        wait_for: Iterable[PrefectFuture[Any]] | None = None,
        dependencies: dict[str, set[RunInput]] | None = None,
    ) -> PrefectConcurrentFuture[R]:
        """
        Submit a task to the task run engine running in a separate thread.

        Args:
            task: The task to submit.
            parameters: The parameters to use when running the task.
            wait_for: A list of futures that the task depends on.

        Returns:
            A future object that can be used to wait for the task to complete and
            retrieve the result.
        """
        if not self._started or self._executor is None:
            raise RuntimeError("Task runner is not started")

        if wait_for and task.tags and (self._max_workers <= len(task.tags)):
            self.logger.warning(
                f"Task {task.name} has {len(task.tags)} tags but only {self._max_workers} workers available"
                "This may lead to dead-locks. Consider increasing the value of `PREFECT_TASK_RUNNER_THREAD_POOL_MAX_WORKERS` or `max_workers`."
            )

        from prefect.context import FlowRunContext
        from prefect.task_engine import run_task_async, run_task_sync

        task_run_id = uuid7()
        cancel_event = threading.Event()
        self._cancel_events[task_run_id] = cancel_event
        context = copy_context()

        flow_run_ctx = FlowRunContext.get()
        if flow_run_ctx:
            get_run_logger(flow_run_ctx).debug(
                f"Submitting task {task.name} to thread pool executor..."
            )
        else:
            self.logger.debug(f"Submitting task {task.name} to thread pool executor...")

        submit_kwargs: dict[str, Any] = dict(
            task=task,
            task_run_id=task_run_id,
            parameters=parameters,
            wait_for=wait_for,
            return_type="state",
            dependencies=dependencies,
            context=dict(cancel_event=cancel_event),
        )

        if task.isasync:
            # TODO: Explore possibly using a long-lived thread with an event loop
            # for better performance
            future = self._executor.submit(
                context.run,
                asyncio.run,
                run_task_async(**submit_kwargs),
            )
        else:
            future = self._executor.submit(
                context.run,
                run_task_sync,
                **submit_kwargs,
            )
        prefect_future = PrefectConcurrentFuture(
            task_run_id=task_run_id, wrapped_future=future
        )
        return prefect_future

    @overload
    def map(
        self,
        task: "Task[P, CoroutineType[Any, Any, R]]",
        parameters: dict[str, Any],
        wait_for: Iterable[PrefectFuture[Any]] | None = None,
    ) -> PrefectFutureList[PrefectConcurrentFuture[R]]: ...

    @overload
    def map(
        self,
        task: "Task[Any, R]",
        parameters: dict[str, Any],
        wait_for: Iterable[PrefectFuture[Any]] | None = None,
    ) -> PrefectFutureList[PrefectConcurrentFuture[R]]: ...

    def map(
        self,
        task: "Task[P, R | CoroutineType[Any, Any, R]]",
        parameters: dict[str, Any],
        wait_for: Iterable[PrefectFuture[Any]] | None = None,
    ) -> PrefectFutureList[PrefectConcurrentFuture[R]]:
        return super().map(task, parameters, wait_for)

    def cancel_all(self) -> None:
        for event in self._cancel_events.values():
            event.set()
            self.logger.debug("Set cancel event")

        if self._executor is not None:
            self._executor.shutdown(cancel_futures=True)
            self._executor = None

    def __enter__(self) -> Self:
        super().__enter__()
        self._executor = ThreadPoolExecutor(max_workers=self._max_workers)
        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        self.cancel_all()
        if self._executor is not None:
            self._executor.shutdown(cancel_futures=True)
            self._executor = None
        super().__exit__(exc_type, exc_value, traceback)

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, ThreadPoolTaskRunner):
            return False
        return self._max_workers == value._max_workers


# Here, we alias ConcurrentTaskRunner to ThreadPoolTaskRunner for backwards compatibility
ConcurrentTaskRunner = ThreadPoolTaskRunner

_PROCESS_POOL_MESSAGE_TYPE_EVENT = "event"
_PROCESS_POOL_MESSAGE_TYPE_LOG = "log"
_PROCESS_POOL_MESSAGE_QUEUE_SHUTDOWN = "__prefect_process_pool_message_queue_shutdown__"

_SubprocessMessageProcessorResult = tuple[str, Any] | None


class _SubprocessMessageProcessor(Protocol):
    def __call__(
        self, message_type: str, message_payload: Any
    ) -> _SubprocessMessageProcessorResult: ...


class _SubprocessMessageProcessorFactory(Protocol):
    def __call__(self) -> _SubprocessMessageProcessor: ...


def _enqueue_process_pool_log(message_queue: Any, log_payload: dict[str, Any]) -> None:
    message_queue.put((_PROCESS_POOL_MESSAGE_TYPE_LOG, log_payload))


def _initialize_process_pool_worker(message_queue: Any | None = None) -> None:
    """
    Configure process-pool workers to forward emitted events and API logs back to
    the parent process.
    """
    if message_queue is None:
        EventsWorker.set_client_override(None)
        set_api_log_sink(None)
        return

    EventsWorker.set_client_override(
        ProcessPoolForwardingEventsClient,
        event_queue=message_queue,
        item_type=_PROCESS_POOL_MESSAGE_TYPE_EVENT,
    )
    set_api_log_sink(partial(_enqueue_process_pool_log, message_queue))


def _run_task_in_subprocess(
    *args: Any,
    env: dict[str, str] | None = None,
    **kwargs: Any,
) -> Any:
    """
    Wrapper function to update environment variables and settings before running a task in a subprocess.
    """
    from prefect.context import hydrated_context
    from prefect.engine import handle_engine_signals
    from prefect.task_engine import run_task_async, run_task_sync

    # Update environment variables
    os.environ.update(env or {})

    # Extract context from kwargs
    context = kwargs.pop("context", None)

    with hydrated_context(context):
        with handle_engine_signals(kwargs.get("task_run_id")):
            # Determine if this is an async task
            task = kwargs.get("task")
            if task and task.isasync:
                # For async tasks, we need to create a new event loop
                import asyncio

                maybe_coro = run_task_async(*args, **kwargs)
                result = asyncio.run(maybe_coro)
            else:
                result = run_task_sync(*args, **kwargs)

            # Flush the subprocess's EventsWorker so that any events emitted
            # during the task are forwarded to the parent process's
            # multiprocessing queue before this function returns.  Without
            # this, the process pool may shut down the worker before the
            # async EventsWorker processing completes, causing events to be
            # silently lost.
            #
            # Only flush when a worker already exists (i.e. events were
            # actually emitted) to avoid instantiating a new EventsWorker
            # and its associated orchestration client for tasks that never
            # emit events.
            with EventsWorker._instance_lock:
                has_events_worker = bool(EventsWorker._instances)
            if has_events_worker:
                try:
                    EventsWorker.instance().wait_until_empty()
                except Exception:
                    get_logger("task_runner").debug(
                        "Failed to flush subprocess EventsWorker; "
                        "some events may not have been forwarded to the parent"
                        " process",
                        exc_info=True,
                    )

            return result


class _ChainedFuture(concurrent.futures.Future[bytes]):
    """Wraps a future-of-future and unwraps the result."""

    def __init__(
        self,
        resolution_future: concurrent.futures.Future[concurrent.futures.Future[bytes]],
    ):
        super().__init__()
        self._resolution_future = resolution_future
        self._process_future: concurrent.futures.Future[bytes] | None = None

        # When resolution completes, hook up to the process future
        def on_resolution_done(
            fut: concurrent.futures.Future[concurrent.futures.Future[bytes]],
        ) -> None:
            try:
                self._process_future = fut.result()

                # Forward process future result to this future
                def on_process_done(
                    process_fut: concurrent.futures.Future[bytes],
                ) -> None:
                    try:
                        result = process_fut.result()
                        self.set_result(result)
                    except Exception as e:
                        self.set_exception(e)

                self._process_future.add_done_callback(on_process_done)
            except Exception as e:
                self.set_exception(e)

        resolution_future.add_done_callback(on_resolution_done)

    def cancel(self) -> bool:
        if self._process_future:
            return self._process_future.cancel()
        return self._resolution_future.cancel()

    def cancelled(self) -> bool:
        if self._process_future:
            return self._process_future.cancelled()
        return self._resolution_future.cancelled()


class _UnpicklingFuture(concurrent.futures.Future[R]):
    """Wrapper for a Future that unpickles the result returned by cloudpickle_wrapped_call."""

    def __init__(self, wrapped_future: concurrent.futures.Future[bytes]):
        self.wrapped_future = wrapped_future

    def result(self, timeout: float | None = None) -> R:
        pickled_result = self.wrapped_future.result(timeout)
        import cloudpickle

        return cloudpickle.loads(pickled_result)

    def exception(self, timeout: float | None = None) -> BaseException | None:
        return self.wrapped_future.exception(timeout)

    def done(self) -> bool:
        return self.wrapped_future.done()

    def cancelled(self) -> bool:
        return self.wrapped_future.cancelled()

    def cancel(self) -> bool:
        return self.wrapped_future.cancel()

    def add_done_callback(
        self, fn: Callable[[concurrent.futures.Future[R]], object]
    ) -> None:
        def _fn(wrapped_future: concurrent.futures.Future[bytes]) -> None:
            import cloudpickle

            result = cloudpickle.loads(wrapped_future.result())
            fn(result)

        return self.wrapped_future.add_done_callback(_fn)


class ProcessPoolTaskRunner(TaskRunner[PrefectConcurrentFuture[Any]]):
    """
    A task runner that executes tasks in a separate process pool.

    This task runner uses `ProcessPoolExecutor` to run tasks in separate processes,
    providing true parallelism for CPU-bound tasks and process isolation. Tasks
    are executed with proper context propagation and error handling.

    Attributes:
        max_workers: The maximum number of processes to use for executing tasks.
            Defaults to `multiprocessing.cpu_count()` if `PREFECT_TASKS_RUNNER_PROCESS_POOL_MAX_WORKERS` is not set.

    Examples:
        Use a process pool task runner with a flow:

        ```python
        from prefect import flow, task
        from prefect.task_runners import ProcessPoolTaskRunner

        @task
        def compute_heavy_task(n: int) -> int:
            # CPU-intensive computation that benefits from process isolation
            return sum(i ** 2 for i in range(n))

        @flow(task_runner=ProcessPoolTaskRunner(max_workers=4))
        def my_flow():
            futures = []
            for i in range(10):
                future = compute_heavy_task.submit(i * 1000)
                futures.append(future)

            return [future.result() for future in futures]
        ```

        Use a process pool task runner as a context manager:

        ```python
        from prefect.task_runners import ProcessPoolTaskRunner

        @task
        def my_task(x: int) -> int:
            return x * 2

        # Use the runner directly
        with ProcessPoolTaskRunner(max_workers=2) as runner:
            future1 = runner.submit(my_task, {"x": 1})
            future2 = runner.submit(my_task, {"x": 2})

            result1 = future1.result()  # 2
            result2 = future2.result()  # 4
        ```

        Configure max workers via settings:

        ```python
        # Set via environment variable
        # export PREFECT_TASKS_RUNNER_PROCESS_POOL_MAX_WORKERS=8

        from prefect import flow
        from prefect.task_runners import ProcessPoolTaskRunner

        @flow(task_runner=ProcessPoolTaskRunner())  # Uses 8 workers from setting
        def my_flow():
            ...
        ```

    Note:
        Process pool task runners provide process isolation but have overhead for
        inter-process communication. They are most beneficial for CPU-bound tasks
        that can take advantage of multiple CPU cores. For I/O-bound tasks,
        consider using `ThreadPoolTaskRunner` instead.

        This runner uses the 'spawn' multiprocessing start method for cross-platform
        consistency and to avoid issues with shared state between processes.

        All task parameters and return values must be serializable with cloudpickle.
        The runner automatically handles context propagation and environment
        variable passing to subprocess workers.
    """

    def __init__(
        self,
        max_workers: int | None = None,
        subprocess_message_processor_factories: (
            Iterable[_SubprocessMessageProcessorFactory] | None
        ) = None,
    ):
        super().__init__()
        current_settings = get_current_settings()
        self._executor: ProcessPoolExecutor | None = None
        self._resolver_executor: ThreadPoolExecutor | None = None
        self._max_workers = (
            max_workers
            or current_settings.tasks.runner.process_pool_max_workers
            or multiprocessing.cpu_count()
        )
        self._cancel_events: dict[uuid.UUID, multiprocessing.Event] = {}
        self._subprocess_message_queue: Any | None = None
        self._message_forwarding_thread: threading.Thread | None = None
        self._subprocess_message_processor_factories: tuple[
            _SubprocessMessageProcessorFactory, ...
        ] = tuple(subprocess_message_processor_factories or ())
        self._cached_context: dict[str, Any] | None = None
        self._cached_env: dict[str, str] | None = None

    def duplicate(self) -> Self:
        duplicate_runner = type(self)(max_workers=self._max_workers)
        duplicate_runner.subprocess_message_processor_factories = (
            self._subprocess_message_processor_factories
        )
        return duplicate_runner

    @property
    def subprocess_message_processor_factories(
        self,
    ) -> tuple[_SubprocessMessageProcessorFactory, ...]:
        return self._subprocess_message_processor_factories

    @subprocess_message_processor_factories.setter
    def subprocess_message_processor_factories(
        self,
        subprocess_message_processor_factories: (
            Iterable[_SubprocessMessageProcessorFactory] | None
        ) = None,
    ) -> None:
        if self._started:
            raise RuntimeError(
                "Cannot configure subprocess message processor factories while task runner is started"
            )
        self._subprocess_message_processor_factories = tuple(
            subprocess_message_processor_factories or ()
        )

    def set_subprocess_message_processor_factories(
        self,
        subprocess_message_processor_factories: (
            Iterable[_SubprocessMessageProcessorFactory] | None
        ) = None,
    ) -> None:
        self.subprocess_message_processor_factories = (
            subprocess_message_processor_factories
        )

    def _process_subprocess_message(
        self,
        message_type: str,
        message_payload: Any,
        message_processors: tuple[_SubprocessMessageProcessor, ...],
    ) -> tuple[str, Any] | None:
        processed_message_type = message_type
        processed_message_payload = message_payload

        for processor in message_processors:
            try:
                processed_message = processor(
                    processed_message_type, processed_message_payload
                )
            except Exception:
                self.logger.exception(
                    "Dropping subprocess message because processor %r raised an exception.",
                    processor,
                )
                return None

            if processed_message is None:
                return None
            if (
                not isinstance(processed_message, tuple)
                or len(processed_message) != 2
                or not isinstance(processed_message[0], str)
            ):
                self.logger.warning(
                    "Dropping subprocess message because processor %r returned invalid payload: %r",
                    processor,
                    processed_message,
                )
                return None

            processed_message_type, processed_message_payload = processed_message

        return processed_message_type, processed_message_payload

    def _forward_subprocess_messages(self) -> None:
        message_queue = self._subprocess_message_queue
        if message_queue is None:
            return

        events_worker: EventsWorker | None = None
        api_log_worker: APILogWorker | None = None
        disable_event_forwarding = False
        disable_log_forwarding = False
        message_processors: list[_SubprocessMessageProcessor] = []
        for processor_factory in self._subprocess_message_processor_factories:
            try:
                message_processors.append(processor_factory())
            except Exception:
                self.logger.exception(
                    "Ignoring subprocess message processor factory %r after initialization failure.",
                    processor_factory,
                )
        message_processor_tuple = tuple(message_processors)

        while True:
            try:
                queued_item = message_queue.get()
            except (EOFError, OSError):
                return

            if queued_item == _PROCESS_POOL_MESSAGE_QUEUE_SHUTDOWN:
                return

            if (
                not isinstance(queued_item, tuple)
                or len(queued_item) != 2
                or not isinstance(queued_item[0], str)
            ):
                self.logger.warning(
                    "Ignoring unexpected subprocess message payload: %r",
                    queued_item,
                )
                continue

            message_type, message_payload = queued_item
            processed_message = self._process_subprocess_message(
                message_type,
                message_payload,
                message_processor_tuple,
            )
            if processed_message is None:
                continue
            message_type, message_payload = processed_message

            try:
                if message_type == _PROCESS_POOL_MESSAGE_TYPE_EVENT:
                    if not isinstance(message_payload, Event):
                        self.logger.warning(
                            "Ignoring unexpected subprocess event payload type: %r",
                            type(message_payload),
                        )
                        continue
                    if disable_event_forwarding:
                        continue
                    if events_worker is None:
                        events_worker = EventsWorker.instance()
                    events_worker.send(message_payload)
                elif message_type == _PROCESS_POOL_MESSAGE_TYPE_LOG:
                    if not isinstance(message_payload, dict):
                        self.logger.warning(
                            "Ignoring unexpected subprocess log payload type: %r",
                            type(message_payload),
                        )
                        continue
                    if disable_log_forwarding:
                        continue
                    if api_log_worker is None:
                        api_log_worker = APILogWorker.instance()
                    api_log_worker.send(message_payload)
                else:
                    self.logger.warning(
                        "Ignoring unknown subprocess message type: %r", message_type
                    )
            except RuntimeError:
                if message_type == _PROCESS_POOL_MESSAGE_TYPE_EVENT:
                    disable_event_forwarding = True
                    events_worker = None
                elif message_type == _PROCESS_POOL_MESSAGE_TYPE_LOG:
                    disable_log_forwarding = True
                    api_log_worker = None

                self.logger.debug(
                    "Disabling subprocess %s forwarding after worker runtime error.",
                    message_type,
                    exc_info=True,
                )
            except Exception:
                self.logger.exception(
                    "Failed to forward subprocess-emitted message to parent worker."
                )

    def _stop_message_forwarding(self) -> None:
        message_queue = self._subprocess_message_queue
        forwarding_thread = self._message_forwarding_thread

        if message_queue is not None:
            try:
                message_queue.put_nowait(_PROCESS_POOL_MESSAGE_QUEUE_SHUTDOWN)
            except (ValueError, OSError):
                pass

        if forwarding_thread is not None:
            forwarding_thread.join(timeout=5)
            if forwarding_thread.is_alive():
                self.logger.warning(
                    "Timed out waiting for process-pool message forwarding to stop."
                )
                # Leave references intact so we don't race with a live thread.
                return
            self._message_forwarding_thread = None

        if message_queue is not None:
            try:
                message_queue.close()
                message_queue.join_thread()
            except (AttributeError, OSError, ValueError):
                pass
            self._subprocess_message_queue = None

    def _resolve_futures_and_submit(
        self,
        task: "Task[P, R | CoroutineType[Any, Any, R]]",
        task_run_id: uuid.UUID,
        parameters: dict[str, Any],
        wait_for: Iterable[PrefectFuture[Any]] | None,
        dependencies: dict[str, set[RunInput]] | None,
        context: dict[str, Any],
        env: dict[str, str],
    ) -> concurrent.futures.Future[bytes]:
        """
        Helper method that:
        1. Waits for all futures in wait_for to complete
        2. Resolves any futures in parameters to their actual values
        3. Submits the task to the ProcessPoolExecutor with resolved values

        This method runs in a background thread to keep submit() non-blocking.
        """
        from prefect.utilities.engine import resolve_inputs_sync

        # Wait for all futures in wait_for to complete
        if wait_for:
            wait(list(wait_for))

        # Resolve any futures in parameters to their actual values
        resolved_parameters = resolve_inputs_sync(
            parameters, return_data=True, max_depth=-1
        )

        # Now submit to the process pool with resolved values
        submit_kwargs: dict[str, Any] = dict(
            task=task,
            task_run_id=task_run_id,
            parameters=resolved_parameters,
            wait_for=None,  # Already waited, no need to pass futures to subprocess
            return_type="state",
            dependencies=dependencies,
            context=context,
        )

        # Prepare the cloudpickle wrapped call for subprocess execution
        wrapped_call = cloudpickle_wrapped_call(
            _run_task_in_subprocess,
            env=env,
            **submit_kwargs,
        )

        # Submit to executor
        return self._executor.submit(wrapped_call)

    @overload
    def submit(
        self,
        task: "Task[P, CoroutineType[Any, Any, R]]",
        parameters: dict[str, Any],
        wait_for: Iterable[PrefectFuture[Any]] | None = None,
        dependencies: dict[str, set[RunInput]] | None = None,
    ) -> PrefectConcurrentFuture[R]: ...

    @overload
    def submit(
        self,
        task: "Task[Any, R]",
        parameters: dict[str, Any],
        wait_for: Iterable[PrefectFuture[Any]] | None = None,
        dependencies: dict[str, set[RunInput]] | None = None,
    ) -> PrefectConcurrentFuture[R]: ...

    def submit(
        self,
        task: "Task[P, R | CoroutineType[Any, Any, R]]",
        parameters: dict[str, Any],
        wait_for: Iterable[PrefectFuture[Any]] | None = None,
        dependencies: dict[str, set[RunInput]] | None = None,
    ) -> PrefectConcurrentFuture[R]:
        """
        Submit a task to the task run engine running in a separate process.

        Args:
            task: The task to submit.
            parameters: The parameters to use when running the task.
            wait_for: A list of futures that the task depends on.
            dependencies: A dictionary of dependencies for the task.

        Returns:
            A future object that can be used to wait for the task to complete and
            retrieve the result.
        """
        if (
            not self._started
            or self._executor is None
            or self._resolver_executor is None
        ):
            raise RuntimeError("Task runner is not started")

        if wait_for and task.tags and (self._max_workers <= len(task.tags)):
            self.logger.warning(
                f"Task {task.name} has {len(task.tags)} tags but only {self._max_workers} workers available"
                "This may lead to dead-locks. Consider increasing the value of `PREFECT_TASKS_RUNNER_PROCESS_POOL_MAX_WORKERS` or `max_workers`."
            )

        from prefect.context import FlowRunContext

        task_run_id = uuid7()

        flow_run_ctx = FlowRunContext.get()
        if flow_run_ctx:
            get_run_logger(flow_run_ctx).debug(
                f"Submitting task {task.name} to process pool executor..."
            )
        else:
            self.logger.debug(
                f"Submitting task {task.name} to process pool executor..."
            )

        # Serialize the current context for the subprocess (cached per runner lifecycle)
        if self._cached_context is None:
            from prefect.context import serialize_context

            self._cached_context = serialize_context()
            self._cached_env = (
                get_current_settings().to_environment_variables(exclude_unset=True)
                | os.environ
            )
        context = self._cached_context
        env = self._cached_env

        # Submit the resolution and subprocess execution to a background thread
        # This keeps submit() non-blocking while still resolving futures before pickling
        resolution_future = self._resolver_executor.submit(
            self._resolve_futures_and_submit,
            task=task,
            task_run_id=task_run_id,
            parameters=parameters,
            wait_for=wait_for,
            dependencies=dependencies,
            context=context,
            env=env,
        )

        # Create a future that chains: thread resolution -> process execution -> unpickling
        # We need to wrap the resolution_future's result (which will be a process future)
        chained_future = _ChainedFuture(resolution_future)

        # Create a PrefectConcurrentFuture that handles unpickling
        prefect_future: PrefectConcurrentFuture[R] = PrefectConcurrentFuture(
            task_run_id=task_run_id, wrapped_future=_UnpicklingFuture(chained_future)
        )
        return prefect_future

    @overload
    def map(
        self,
        task: "Task[P, CoroutineType[Any, Any, R]]",
        parameters: dict[str, Any],
        wait_for: Iterable[PrefectFuture[Any]] | None = None,
    ) -> PrefectFutureList[PrefectConcurrentFuture[R]]: ...

    @overload
    def map(
        self,
        task: "Task[Any, R]",
        parameters: dict[str, Any],
        wait_for: Iterable[PrefectFuture[Any]] | None = None,
    ) -> PrefectFutureList[PrefectConcurrentFuture[R]]: ...

    def map(
        self,
        task: "Task[P, R | CoroutineType[Any, Any, R]]",
        parameters: dict[str, Any],
        wait_for: Iterable[PrefectFuture[Any]] | None = None,
    ) -> PrefectFutureList[PrefectConcurrentFuture[R]]:
        return super().map(task, parameters, wait_for)

    def cancel_all(self) -> None:
        # Invalidate cached context and env so they are recomputed on next start
        self._cached_context = None
        self._cached_env = None

        # Clear cancel events first to avoid resource tracking issues
        events_to_set = list(self._cancel_events.values())
        self._cancel_events.clear()

        for event in events_to_set:
            try:
                event.set()
                self.logger.debug("Set cancel event")
            except (OSError, ValueError):
                # Ignore errors if event is already closed/invalid
                pass

        if self._executor is not None:
            self._executor.shutdown(cancel_futures=True, wait=True)
            self._executor = None

        self._stop_message_forwarding()

    def __enter__(self) -> Self:
        super().__enter__()
        # Use spawn method for cross-platform consistency and avoiding shared state issues
        mp_context = multiprocessing.get_context("spawn")
        self._subprocess_message_queue = mp_context.Queue()
        self._message_forwarding_thread = threading.Thread(
            target=self._forward_subprocess_messages,
            name="ProcessPoolTaskRunnerMessageForwarder",
            daemon=True,
        )
        self._message_forwarding_thread.start()

        try:
            self._executor = ProcessPoolExecutor(
                max_workers=self._max_workers,
                mp_context=mp_context,
                initializer=_initialize_process_pool_worker,
                initargs=(self._subprocess_message_queue,),
            )
            # Create a thread pool for resolving futures before submitting to process pool
            self._resolver_executor = ThreadPoolExecutor(max_workers=self._max_workers)
        except Exception:
            if self._executor is not None:
                self._executor.shutdown(cancel_futures=True, wait=True)
                self._executor = None
            self._stop_message_forwarding()
            raise

        return self

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        self.cancel_all()
        # cancel_all() already shuts down the executor, but double-check
        if self._executor is not None:
            self._executor.shutdown(cancel_futures=True, wait=True)
            self._executor = None
        if self._resolver_executor is not None:
            self._resolver_executor.shutdown(cancel_futures=True, wait=True)
            self._resolver_executor = None
        super().__exit__(exc_type, exc_value, traceback)

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, ProcessPoolTaskRunner):
            return False
        return self._max_workers == value._max_workers


class PrefectTaskRunner(TaskRunner[PrefectDistributedFuture[R]]):
    def __init__(self):
        super().__init__()

    def duplicate(self) -> "PrefectTaskRunner[R]":
        return type(self)()

    @overload
    def submit(
        self,
        task: "Task[P, CoroutineType[Any, Any, R]]",
        parameters: dict[str, Any],
        wait_for: Iterable[PrefectFuture[Any]] | None = None,
        dependencies: dict[str, set[RunInput]] | None = None,
    ) -> PrefectDistributedFuture[R]: ...

    @overload
    def submit(
        self,
        task: "Task[Any, R]",
        parameters: dict[str, Any],
        wait_for: Iterable[PrefectFuture[Any]] | None = None,
        dependencies: dict[str, set[RunInput]] | None = None,
    ) -> PrefectDistributedFuture[R]: ...

    def submit(
        self,
        task: "Task[P, R | CoroutineType[Any, Any, R]]",
        parameters: dict[str, Any],
        wait_for: Iterable[PrefectFuture[Any]] | None = None,
        dependencies: dict[str, set[RunInput]] | None = None,
    ) -> PrefectDistributedFuture[R]:
        """
        Submit a task to the task run engine running in a separate thread.

        Args:
            task: The task to submit.
            parameters: The parameters to use when running the task.
            wait_for: A list of futures that the task depends on.

        Returns:
            A future object that can be used to wait for the task to complete and
            retrieve the result.
        """
        if not self._started:
            raise RuntimeError("Task runner is not started")
        from prefect.context import FlowRunContext

        flow_run_ctx = FlowRunContext.get()
        if flow_run_ctx:
            get_run_logger(flow_run_ctx).info(
                f"Submitting task {task.name} to for execution by a Prefect task worker..."
            )
        else:
            self.logger.info(
                f"Submitting task {task.name} to for execution by a Prefect task worker..."
            )

        return task.apply_async(
            kwargs=parameters, wait_for=wait_for, dependencies=dependencies
        )

    @overload
    def map(
        self,
        task: "Task[P, CoroutineType[Any, Any, R]]",
        parameters: dict[str, Any],
        wait_for: Iterable[PrefectFuture[Any]] | None = None,
    ) -> PrefectFutureList[PrefectDistributedFuture[R]]: ...

    @overload
    def map(
        self,
        task: "Task[Any, R]",
        parameters: dict[str, Any],
        wait_for: Iterable[PrefectFuture[Any]] | None = None,
    ) -> PrefectFutureList[PrefectDistributedFuture[R]]: ...

    def map(
        self,
        task: "Task[P, R | CoroutineType[Any, Any, R]]",
        parameters: dict[str, Any],
        wait_for: Iterable[PrefectFuture[Any]] | None = None,
    ) -> PrefectFutureList[PrefectDistributedFuture[R]]:
        return super().map(task, parameters, wait_for)



"""
Utilities for managing worker threads.
"""

from __future__ import annotations

import asyncio
import atexit
import concurrent.futures
import itertools
import os
import queue
import threading
import weakref
from typing import Any, Optional

from typing_extensions import TypeVar

from prefect._internal.concurrency import logger
from prefect._internal.concurrency.calls import Call, Portal
from prefect._internal.concurrency.cancellation import CancelledError
from prefect._internal.concurrency.event_loop import get_running_loop
from prefect._internal.concurrency.primitives import Event

T = TypeVar("T", infer_variance=True)

# Track all active instances for fork handling
_active_instances: weakref.WeakSet[WorkerThread | EventLoopThread] = weakref.WeakSet()


def _reset_after_fork_in_child():
    """
    Reset thread state after fork() to prevent multiprocessing deadlocks on Linux.

    When fork() is called, the child process inherits all thread state and locks
    from the parent, but only the calling thread continues. This leaves other threads'
    locks in inconsistent states causing deadlocks.

    This handler is called by os.register_at_fork() in the child process after fork().
    """
    for instance in list(_active_instances):
        instance.reset_for_fork()


# Register fork handler if supported (POSIX systems)
if hasattr(os, "register_at_fork"):
    try:
        os.register_at_fork(after_in_child=_reset_after_fork_in_child)
    except RuntimeError:
        # Might fail in certain contexts (e.g., if already in a child process)
        pass


class WorkerThread(Portal):
    """
    A portal to a worker running on a thread.
    """

    # Used for unique thread names by default
    _counter = itertools.count().__next__

    def __init__(
        self, name: Optional[str] = None, daemon: bool = False, run_once: bool = False
    ):
        name = name or f"WorkerThread-{self._counter()}"

        self.thread = threading.Thread(
            name=name, daemon=daemon, target=self._entrypoint
        )
        self._queue: queue.Queue[Optional[Call[Any]]] = queue.Queue()
        self._run_once: bool = run_once
        self._started: bool = False
        self._submitted_count: int = 0
        self._lock = threading.Lock()

        # Track this instance for fork handling
        _active_instances.add(self)

        if not daemon:
            atexit.register(self.shutdown)

    def reset_for_fork(self) -> None:
        """Reset state after fork() to prevent deadlocks in child process."""
        self._started = False
        self._queue = queue.Queue()
        self._lock = threading.Lock()
        self._submitted_count = 0

    def start(self) -> None:
        """
        Start the worker thread.
        """
        with self._lock:
            if not self._started:
                self._started = True
                self.thread.start()

    def submit(self, call: Call[T]) -> Call[T]:
        if self._submitted_count > 0 and self._run_once:
            raise RuntimeError(
                "Worker configured to only run once. A call has already been submitted."
            )

        # Start on first submission if not started
        if not self._started:
            self.start()

        # Track the portal running the call
        call.set_runner(self)

        # Put the call in the queue
        self._queue.put_nowait(call)

        self._submitted_count += 1
        if self._run_once:
            call.future.add_done_callback(lambda _: self.shutdown())

        return call

    def shutdown(self) -> None:
        """
        Shutdown the worker thread. Does not wait for the thread to stop.
        """
        self._queue.put_nowait(None)

    @property
    def name(self) -> str:
        return self.thread.name

    def _entrypoint(self) -> None:
        """
        Entrypoint for the thread.
        """
        try:
            self._run_until_shutdown()
        except CancelledError:
            logger.exception("%s was cancelled", self.name)
        except BaseException:
            # Log exceptions that crash the thread
            logger.exception("%s encountered exception", self.name)
            raise

    def _run_until_shutdown(self):
        while True:
            call = self._queue.get()
            if call is None:
                logger.info("Exiting worker thread %r", self.name)
                break  # shutdown requested

            task = call.run()
            assert task is None  # calls should never return a coroutine in this worker
            del call

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *_):
        self.shutdown()


class EventLoopThread(Portal):
    """
    A portal to a worker running on a thread with an event loop.
    """

    def __init__(
        self,
        name: str = "EventLoopThread",
        daemon: bool = False,
        run_once: bool = False,
    ):
        self.thread = threading.Thread(
            name=name, daemon=daemon, target=self._entrypoint
        )
        self._ready_future: concurrent.futures.Future[bool] = (
            concurrent.futures.Future()
        )
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._shutdown_event: Event = Event()
        self._run_once: bool = run_once
        self._submitted_count: int = 0
        self._on_shutdown: list[Call[Any]] = []
        self._lock = threading.Lock()

        # Track this instance for fork handling
        _active_instances.add(self)

        if not daemon:
            atexit.register(self.shutdown)

    def reset_for_fork(self) -> None:
        """Reset state after fork() to prevent deadlocks in child process."""
        self._loop = None
        self._ready_future = concurrent.futures.Future()
        self._shutdown_event = Event()
        self._lock = threading.Lock()
        self._submitted_count = 0
        self._on_shutdown = []

    def start(self):
        """
        Start the worker thread; raises any exceptions encountered during startup.
        """
        with self._lock:
            if self._loop is None:
                self.thread.start()
                self._ready_future.result()

    def submit(self, call: Call[T]) -> Call[T]:
        if self._loop is None:
            self.start()

        with self._lock:
            if self._submitted_count > 0 and self._run_once:
                raise RuntimeError(
                    "Worker configured to only run once. A call has already been"
                    " submitted."
                )

            if self._shutdown_event.is_set():
                raise RuntimeError("Worker is shutdown.")

            # Track the portal running the call
            call.set_runner(self)

            if self._run_once:
                call.future.add_done_callback(lambda _: self.shutdown())

            # Submit the call to the event loop
            assert self._loop is not None
            asyncio.run_coroutine_threadsafe(self._run_call(call), self._loop)
            self._submitted_count += 1

        return call

    def shutdown(self) -> None:
        """
        Shutdown the worker thread. Does not wait for the thread to stop.
        """
        with self._lock:
            self._shutdown_event.set()

    @property
    def name(self) -> str:
        return self.thread.name

    @property
    def running(self) -> bool:
        return not self._shutdown_event.is_set()

    @property
    def loop(self) -> asyncio.AbstractEventLoop | None:
        return self._loop

    def _entrypoint(self):
        """
        Entrypoint for the thread.

        Immediately create a new event loop and pass control to `run_until_shutdown`.
        """
        try:
            asyncio.run(self._run_until_shutdown())
        except BaseException:
            # Log exceptions that crash the thread
            logger.exception("%s encountered exception", self.name)
            raise

    async def _run_until_shutdown(self):
        try:
            self._loop = asyncio.get_running_loop()
            self._ready_future.set_result(True)
        except Exception as exc:
            self._ready_future.set_exception(exc)
            return

        await self._shutdown_event.wait()

        for call in self._on_shutdown:
            await self._run_call(call)

        # Empty the list to allow calls to be garbage collected. Issue #10338.
        self._on_shutdown = []

    async def _run_call(self, call: Call[Any]) -> None:
        task = call.run()
        if task is not None:
            await task

    def add_shutdown_call(self, call: Call[Any]) -> None:
        self._on_shutdown.append(call)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *_):
        self.shutdown()


# the GLOBAL LOOP is used for background services, like logs
_global_loop: Optional[EventLoopThread] = None
# the RUN SYNC LOOP is used exclusively for running async functions in a sync context via asyncutils.run_sync
_run_sync_loop: Optional[EventLoopThread] = None


def get_global_loop() -> EventLoopThread:
    """
    Get the global loop thread.

    Creates a new one if there is not one available.
    """
    global _global_loop

    # Create a new worker on first call or if the existing worker is dead
    if (
        _global_loop is None
        or not _global_loop.thread.is_alive()
        or not _global_loop.running
    ):
        _global_loop = EventLoopThread(daemon=True, name="GlobalEventLoopThread")
        _global_loop.start()

    return _global_loop


def in_global_loop() -> bool:
    """
    Check if called from the global loop.
    """
    if _global_loop is None:
        # Avoid creating a global loop if there isn't one
        return False

    return getattr(get_global_loop(), "_loop") == get_running_loop()


def get_run_sync_loop() -> EventLoopThread:
    """
    Get the run_sync loop thread.

    Creates a new one if there is not one available.
    """
    global _run_sync_loop

    # Create a new worker on first call or if the existing worker is dead
    if (
        _run_sync_loop is None
        or not _run_sync_loop.thread.is_alive()
        or not _run_sync_loop.running
    ):
        _run_sync_loop = EventLoopThread(daemon=True, name="RunSyncEventLoopThread")
        _run_sync_loop.start()

    return _run_sync_loop


def in_run_sync_loop() -> bool:
    """
    Check if called from the global loop.
    """
    if _run_sync_loop is None:
        # Avoid creating a global loop if there isn't one
        return False

    return getattr(get_run_sync_loop(), "_loop") == get_running_loop()


def wait_for_global_loop_exit(timeout: Optional[float] = None) -> None:
    """
    Shutdown the global loop and wait for it to exit.
    """
    loop_thread = get_global_loop()
    loop_thread.shutdown()

    if threading.get_ident() == loop_thread.thread.ident:
        raise RuntimeError("Cannot wait for the loop thread from inside itself.")

    loop_thread.thread.join(timeout)



"""
This module contains functions that allow sending type-checked `RunInput` data
to flows at runtime. Flows can send back responses, establishing two-way
channels with senders. These functions are particularly useful for systems that
require ongoing data transfer or need to react to input quickly.
real-time interaction and efficient data handling. It's designed to facilitate
dynamic communication within distributed or microservices-oriented systems,
making it ideal for scenarios requiring continuous data synchronization and
processing. It's particularly useful for systems that require ongoing data
input and output.

The following is an example of two flows. One sends a random number to the
other and waits for a response. The other receives the number, squares it, and
sends the result back. The sender flow then prints the result.

Sender flow:

```python
import random
from uuid import UUID
from prefect import flow
from prefect.logging import get_run_logger
from prefect.input import RunInput

class NumberData(RunInput):
    number: int


@flow
async def sender_flow(receiver_flow_run_id: UUID):
    logger = get_run_logger()

    the_number = random.randint(1, 100)

    await NumberData(number=the_number).send_to(receiver_flow_run_id)

    receiver = NumberData.receive(flow_run_id=receiver_flow_run_id)
    squared = await receiver.next()

    logger.info(f"{the_number} squared is {squared.number}")
```

Receiver flow:
```python
import random
from uuid import UUID
from prefect import flow
from prefect.logging import get_run_logger
from prefect.input import RunInput

class NumberData(RunInput):
    number: int


@flow
async def receiver_flow():
    async for data in NumberData.receive():
        squared = data.number ** 2
        data.respond(NumberData(number=squared))
```
"""

from __future__ import annotations

from inspect import isclass
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Generic,
    Literal,
    Optional,
    Set,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)
from uuid import UUID, uuid4

import anyio
import pydantic
from pydantic import ConfigDict
from typing_extensions import Self

from prefect._internal.compatibility.async_dispatch import async_dispatch
from prefect.input.actions import (
    acreate_flow_run_input,
    acreate_flow_run_input_from_model,
    afilter_flow_run_input,
    aread_flow_run_input,
    create_flow_run_input,
    create_flow_run_input_from_model,
    ensure_flow_run_id,
    filter_flow_run_input,
    read_flow_run_input,
)

if TYPE_CHECKING:
    from prefect.client.schemas.objects import FlowRunInput
    from prefect.states import State

from prefect._internal.pydantic.v2_schema import create_v2_schema, is_v2_model

R = TypeVar("R", bound="RunInput")
T = TypeVar("T", bound="object")

Keyset = Dict[
    Union[Literal["description"], Literal["response"], Literal["schema"]], str
]


def keyset_from_paused_state(state: "State") -> Keyset:
    """
    Get the keyset for the given Paused state.

    Args:
        - state (State): the state to get the keyset for
    """

    if not state.is_paused():
        raise RuntimeError(f"{state.type.value!r} is unsupported.")

    state_name = state.name or ""
    base_key = f"{state_name.lower()}-{str(state.state_details.pause_key)}"
    return keyset_from_base_key(base_key)


def keyset_from_base_key(base_key: str) -> Keyset:
    """
    Get the keyset for the given base key.

    Args:
        - base_key (str): the base key to get the keyset for

    Returns:
        - Dict[str, str]: the keyset
    """
    return {
        "description": f"{base_key}-description",
        "response": f"{base_key}-response",
        "schema": f"{base_key}-schema",
    }


class RunInputMetadata(pydantic.BaseModel):
    key: str
    sender: Optional[str] = None
    receiver: UUID


class BaseRunInput(pydantic.BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(extra="forbid")

    _description: Optional[str] = pydantic.PrivateAttr(default=None)
    _metadata: RunInputMetadata = pydantic.PrivateAttr()

    @property
    def metadata(self) -> RunInputMetadata:
        return self._metadata

    @classmethod
    def keyset_from_type(cls) -> Keyset:
        return keyset_from_base_key(cls.__name__.lower())

    @classmethod
    async def asave(cls, keyset: Keyset, flow_run_id: Optional[UUID] = None) -> None:
        """
        Save the run input response to the given key asynchronously.

        Args:
            - keyset (Keyset): the keyset to save the input for
            - flow_run_id (UUID, optional): the flow run ID to save the input for
        """
        if is_v2_model(cls):
            schema = create_v2_schema(cls.__name__, model_base=cls)
        else:
            schema = cls.model_json_schema(by_alias=True)

        await acreate_flow_run_input(
            key=keyset["schema"], value=schema, flow_run_id=flow_run_id
        )

        description = cls._description if isinstance(cls._description, str) else None
        if description:
            await acreate_flow_run_input(
                key=keyset["description"],
                value=description,
                flow_run_id=flow_run_id,
            )

    @classmethod
    @async_dispatch(asave)
    def save(cls, keyset: Keyset, flow_run_id: Optional[UUID] = None) -> None:
        """
        Save the run input response to the given key.

        Args:
            - keyset (Keyset): the keyset to save the input for
            - flow_run_id (UUID, optional): the flow run ID to save the input for
        """
        if is_v2_model(cls):
            schema = create_v2_schema(cls.__name__, model_base=cls)
        else:
            schema = cls.model_json_schema(by_alias=True)

        create_flow_run_input(
            key=keyset["schema"], value=schema, flow_run_id=flow_run_id
        )

        description = cls._description if isinstance(cls._description, str) else None
        if description:
            create_flow_run_input(
                key=keyset["description"],
                value=description,
                flow_run_id=flow_run_id,
            )

    @classmethod
    async def aload(cls, keyset: Keyset, flow_run_id: Optional[UUID] = None) -> Self:
        """
        Load the run input response from the given key asynchronously.

        Args:
            - keyset (Keyset): the keyset to load the input for
            - flow_run_id (UUID, optional): the flow run ID to load the input for
        """
        flow_run_id = ensure_flow_run_id(flow_run_id)
        value = await aread_flow_run_input(keyset["response"], flow_run_id=flow_run_id)
        if value:
            instance = cls(**value)
        else:
            instance = cls()
        instance._metadata = RunInputMetadata(
            key=keyset["response"], sender=None, receiver=flow_run_id
        )
        return instance

    @classmethod
    @async_dispatch(aload)
    def load(cls, keyset: Keyset, flow_run_id: Optional[UUID] = None) -> Self:
        """
        Load the run input response from the given key.

        Args:
            - keyset (Keyset): the keyset to load the input for
            - flow_run_id (UUID, optional): the flow run ID to load the input for
        """
        flow_run_id = ensure_flow_run_id(flow_run_id)
        value = read_flow_run_input(keyset["response"], flow_run_id=flow_run_id)
        if value:
            instance = cls(**value)
        else:
            instance = cls()
        instance._metadata = RunInputMetadata(
            key=keyset["response"], sender=None, receiver=flow_run_id
        )
        return instance

    @classmethod
    def load_from_flow_run_input(cls, flow_run_input: "FlowRunInput") -> Self:
        """
        Load the run input from a FlowRunInput object.

        Args:
            - flow_run_input (FlowRunInput): the flow run input to load the input for
        """
        instance = cls(**flow_run_input.decoded_value)
        instance._metadata = RunInputMetadata(
            key=flow_run_input.key,
            sender=flow_run_input.sender,
            receiver=flow_run_input.flow_run_id,
        )
        return instance

    @classmethod
    def with_initial_data(
        cls: Type[R], description: Optional[str] = None, **kwargs: Any
    ) -> Type[R]:
        """
        Create a new `RunInput` subclass with the given initial data as field
        defaults.

        Args:
            - description (str, optional): a description to show when resuming
                a flow run that requires input
            - kwargs (Any): the initial data to populate the subclass
        """
        fields: Dict[str, Any] = {}
        for key, value in kwargs.items():
            original_field = cls.model_fields.get(key)
            # Use the original field annotation if it exists, otherwise infer from value
            field_type = original_field.annotation if original_field else type(value)
            fields[key] = (field_type, value)
        model = pydantic.create_model(cls.__name__, **fields, __base__=cls)

        if description is not None:
            model._description = description

        return model

    async def arespond(
        self,
        run_input: "RunInput",
        sender: Optional[str] = None,
        key_prefix: Optional[str] = None,
    ) -> None:
        """Respond to the sender of this input asynchronously."""
        flow_run_id = None
        if self.metadata.sender and self.metadata.sender.startswith("prefect.flow-run"):
            _, _, id = self.metadata.sender.rpartition(".")
            flow_run_id = UUID(id)

        if not flow_run_id:
            raise RuntimeError(
                "Cannot respond to an input that was not sent by a flow run."
            )

        await _asend_input(
            flow_run_id=flow_run_id,
            run_input=run_input,
            sender=sender,
            key_prefix=key_prefix,
        )

    @async_dispatch(arespond)
    def respond(
        self,
        run_input: "RunInput",
        sender: Optional[str] = None,
        key_prefix: Optional[str] = None,
    ) -> None:
        """Respond to the sender of this input."""
        flow_run_id = None
        if self.metadata.sender and self.metadata.sender.startswith("prefect.flow-run"):
            _, _, id = self.metadata.sender.rpartition(".")
            flow_run_id = UUID(id)

        if not flow_run_id:
            raise RuntimeError(
                "Cannot respond to an input that was not sent by a flow run."
            )

        _send_input_sync(
            flow_run_id=flow_run_id,
            run_input=run_input,
            sender=sender,
            key_prefix=key_prefix,
        )

    async def asend_to(
        self,
        flow_run_id: UUID,
        sender: Optional[str] = None,
        key_prefix: Optional[str] = None,
    ) -> None:
        """Send this input to a flow run asynchronously."""
        await _asend_input(
            flow_run_id=flow_run_id,
            run_input=self,
            sender=sender,
            key_prefix=key_prefix,
        )

    @async_dispatch(asend_to)
    def send_to(
        self,
        flow_run_id: UUID,
        sender: Optional[str] = None,
        key_prefix: Optional[str] = None,
    ) -> None:
        """Send this input to a flow run."""
        _send_input_sync(
            flow_run_id=flow_run_id,
            run_input=self,
            sender=sender,
            key_prefix=key_prefix,
        )


class RunInput(BaseRunInput):
    @classmethod
    def receive(
        cls,
        timeout: Optional[float] = 3600,
        poll_interval: float = 10,
        raise_timeout_error: bool = False,
        exclude_keys: Optional[Set[str]] = None,
        key_prefix: Optional[str] = None,
        flow_run_id: Optional[UUID] = None,
    ) -> GetInputHandler[Self]:
        if key_prefix is None:
            key_prefix = f"{cls.__name__.lower()}-auto"

        return GetInputHandler(
            run_input_cls=cls,
            key_prefix=key_prefix,
            timeout=timeout,
            poll_interval=poll_interval,
            raise_timeout_error=raise_timeout_error,
            exclude_keys=exclude_keys,
            flow_run_id=flow_run_id,
        )

    @classmethod
    def subclass_from_base_model_type(
        cls, model_cls: Type[pydantic.BaseModel]
    ) -> Type["RunInput"]:
        """
        Create a new `RunInput` subclass from the given `pydantic.BaseModel`
        subclass.

        Args:
            - model_cls (pydantic.BaseModel subclass): the class from which
                to create the new `RunInput` subclass
        """
        return type(f"{model_cls.__name__}RunInput", (RunInput, model_cls), {})  # type: ignore


class AutomaticRunInput(BaseRunInput, Generic[T]):
    value: T

    @classmethod
    async def aload(cls, keyset: Keyset, flow_run_id: Optional[UUID] = None) -> T:  # type: ignore[override]
        """
        Load the run input response from the given key asynchronously.

        Args:
            - keyset (Keyset): the keyset to load the input for
            - flow_run_id (UUID, optional): the flow run ID to load the input for
        """
        instance = await super().aload(keyset, flow_run_id=flow_run_id)
        return instance.value

    @classmethod
    @async_dispatch(aload)
    def load(cls, keyset: Keyset, flow_run_id: Optional[UUID] = None) -> T:  # type: ignore[override]
        """
        Load the run input response from the given key.

        Args:
            - keyset (Keyset): the keyset to load the input for
            - flow_run_id (UUID, optional): the flow run ID to load the input for
        """
        instance = super().load(keyset, flow_run_id=flow_run_id)
        return instance.value

    @classmethod
    def subclass_from_type(cls, _type: Type[T]) -> Type["AutomaticRunInput[T]"]:
        """
        Create a new `AutomaticRunInput` subclass from the given type.

        This method uses the type's name as a key prefix to identify related
        flow run inputs. This helps in ensuring that values saved under a type
        (like List[int]) are retrievable under the generic type name (like "list").
        """
        fields: Dict[str, Any] = {"value": (_type, ...)}

        # Explanation for using getattr for type name extraction:
        # - "__name__": This is the usual attribute for getting the name of
        #   most types.
        # - "_name": Used as a fallback, some type annotations in Python 3.9
        #   and earlier might only have this attribute instead of __name__.
        # - If neither is available, defaults to an empty string to prevent
        #   errors, but typically we should find at least one valid name
        #   attribute. This will match all automatic inputs sent to the flow
        #   run, rather than a specific type.
        #
        # This approach ensures compatibility across Python versions and
        # handles various edge cases in type annotations.

        type_prefix: str = getattr(
            _type, "__name__", getattr(_type, "_name", "")
        ).lower()

        class_name = f"{type_prefix}AutomaticRunInput"

        # Creating a new Pydantic model class dynamically with the name based
        # on the type prefix.
        new_cls: Type["AutomaticRunInput[T]"] = pydantic.create_model(
            class_name, **fields, __base__=AutomaticRunInput
        )
        return new_cls

    @classmethod
    def receive(
        cls,
        timeout: Optional[float] = 3600,
        poll_interval: float = 10,
        raise_timeout_error: bool = False,
        exclude_keys: Optional[Set[str]] = None,
        key_prefix: Optional[str] = None,
        flow_run_id: Optional[UUID] = None,
        with_metadata: bool = False,
    ) -> GetAutomaticInputHandler[T]:
        key_prefix = key_prefix or f"{cls.__name__.lower()}-auto"

        return GetAutomaticInputHandler(
            run_input_cls=cls,
            key_prefix=key_prefix,
            timeout=timeout,
            poll_interval=poll_interval,
            raise_timeout_error=raise_timeout_error,
            exclude_keys=exclude_keys,
            flow_run_id=flow_run_id,
            with_metadata=with_metadata,
        )


def run_input_subclass_from_type(
    _type: Union[Type[R], Type[T], pydantic.BaseModel],
) -> Union[Type[AutomaticRunInput[T]], Type[R]]:
    """
    Create a new `RunInput` subclass from the given type.
    """
    if isclass(_type):
        if issubclass(_type, RunInput):
            return cast(Type[R], _type)
        elif issubclass(_type, pydantic.BaseModel):
            return cast(Type[R], RunInput.subclass_from_base_model_type(_type))

    # Could be something like a typing._GenericAlias or any other type that
    # isn't a `RunInput` subclass or `pydantic.BaseModel` subclass. Try passing
    # it to AutomaticRunInput to see if we can create a model from it.
    return cast(
        Type[AutomaticRunInput[T]],
        AutomaticRunInput.subclass_from_type(cast(Type[T], _type)),
    )


class GetInputHandler(Generic[R]):
    def __init__(
        self,
        run_input_cls: Type[R],
        key_prefix: str,
        timeout: float | None = 3600,
        poll_interval: float = 10,
        raise_timeout_error: bool = False,
        exclude_keys: Optional[Set[str]] = None,
        flow_run_id: Optional[UUID] = None,
    ):
        self.run_input_cls: Type[R] = run_input_cls
        self.key_prefix: str = key_prefix
        self.timeout: float | None = timeout
        self.poll_interval: float = poll_interval
        self.exclude_keys: set[str] = set()
        self.raise_timeout_error: bool = raise_timeout_error
        self.flow_run_id: UUID = ensure_flow_run_id(flow_run_id)

        if exclude_keys is not None:
            self.exclude_keys.update(exclude_keys)

    def __iter__(self) -> Self:
        return self

    def __next__(self) -> R:
        try:
            return cast(R, self.next())
        except TimeoutError:
            if self.raise_timeout_error:
                raise
            raise StopIteration

    def __aiter__(self) -> Self:
        return self

    async def __anext__(self) -> R:
        try:
            return await self.anext()
        except TimeoutError:
            if self.raise_timeout_error:
                raise
            raise StopAsyncIteration

    async def filter_for_inputs(self) -> list["FlowRunInput"]:
        """Filter for inputs asynchronously."""
        flow_run_inputs = await afilter_flow_run_input(
            key_prefix=self.key_prefix,
            limit=1,
            exclude_keys=self.exclude_keys,
            flow_run_id=self.flow_run_id,
        )

        if flow_run_inputs:
            self.exclude_keys.add(*[i.key for i in flow_run_inputs])

        return flow_run_inputs

    def filter_for_inputs_sync(self) -> list["FlowRunInput"]:
        """Filter for inputs synchronously."""
        flow_run_inputs = filter_flow_run_input(
            key_prefix=self.key_prefix,
            limit=1,
            exclude_keys=self.exclude_keys,
            flow_run_id=self.flow_run_id,
        )

        if flow_run_inputs:
            self.exclude_keys.add(*[i.key for i in flow_run_inputs])

        return flow_run_inputs

    def to_instance(self, flow_run_input: "FlowRunInput") -> R:
        return self.run_input_cls.load_from_flow_run_input(flow_run_input)

    async def anext(self) -> R:
        """Get the next input asynchronously."""
        flow_run_inputs = await self.filter_for_inputs()
        if flow_run_inputs:
            return self.to_instance(flow_run_inputs[0])

        with anyio.fail_after(self.timeout):
            while True:
                await anyio.sleep(self.poll_interval)
                flow_run_inputs = await self.filter_for_inputs()
                if flow_run_inputs:
                    return self.to_instance(flow_run_inputs[0])

    @async_dispatch(anext)
    def next(self) -> R:
        """Get the next input."""
        import time

        flow_run_inputs = self.filter_for_inputs_sync()
        if flow_run_inputs:
            return self.to_instance(flow_run_inputs[0])

        start_time = time.monotonic()
        while True:
            time.sleep(self.poll_interval)
            flow_run_inputs = self.filter_for_inputs_sync()
            if flow_run_inputs:
                return self.to_instance(flow_run_inputs[0])
            if (
                self.timeout is not None
                and time.monotonic() - start_time >= self.timeout
            ):
                raise TimeoutError("Timed out waiting for input")


class GetAutomaticInputHandler(Generic[T]):
    def __init__(
        self,
        run_input_cls: Type[AutomaticRunInput[T]],
        key_prefix: str,
        timeout: float | None = 3600,
        poll_interval: float = 10,
        raise_timeout_error: bool = False,
        exclude_keys: Optional[Set[str]] = None,
        flow_run_id: Optional[UUID] = None,
        with_metadata: bool = False,
    ):
        self.run_input_cls: Type[AutomaticRunInput[T]] = run_input_cls
        self.key_prefix: str = key_prefix
        self.timeout: float | None = timeout
        self.poll_interval: float = poll_interval
        self.exclude_keys: set[str] = set()
        self.raise_timeout_error: bool = raise_timeout_error
        self.flow_run_id: UUID = ensure_flow_run_id(flow_run_id)
        self.with_metadata = with_metadata

        if exclude_keys is not None:
            self.exclude_keys.update(exclude_keys)

    def __iter__(self) -> Self:
        return self

    def __next__(self) -> T | AutomaticRunInput[T]:
        try:
            return self.next()
        except TimeoutError:
            if self.raise_timeout_error:
                raise
            raise StopIteration

    def __aiter__(self) -> Self:
        return self

    async def __anext__(self) -> Union[T, AutomaticRunInput[T]]:
        try:
            return await self.anext()
        except TimeoutError:
            if self.raise_timeout_error:
                raise
            raise StopAsyncIteration

    async def filter_for_inputs(self) -> list["FlowRunInput"]:
        """Filter for inputs asynchronously."""
        flow_run_inputs = await afilter_flow_run_input(
            key_prefix=self.key_prefix,
            limit=1,
            exclude_keys=self.exclude_keys,
            flow_run_id=self.flow_run_id,
        )

        if flow_run_inputs:
            self.exclude_keys.add(*[i.key for i in flow_run_inputs])

        return flow_run_inputs

    def filter_for_inputs_sync(self) -> list["FlowRunInput"]:
        """Filter for inputs synchronously."""
        flow_run_inputs = filter_flow_run_input(
            key_prefix=self.key_prefix,
            limit=1,
            exclude_keys=self.exclude_keys,
            flow_run_id=self.flow_run_id,
        )

        if flow_run_inputs:
            self.exclude_keys.add(*[i.key for i in flow_run_inputs])

        return flow_run_inputs

    async def anext(self) -> Union[T, AutomaticRunInput[T]]:
        """Get the next input asynchronously."""
        flow_run_inputs = await self.filter_for_inputs()
        if flow_run_inputs:
            return self.to_instance(flow_run_inputs[0])

        with anyio.fail_after(self.timeout):
            while True:
                await anyio.sleep(self.poll_interval)
                flow_run_inputs = await self.filter_for_inputs()
                if flow_run_inputs:
                    return self.to_instance(flow_run_inputs[0])

    @async_dispatch(anext)
    def next(self) -> Union[T, AutomaticRunInput[T]]:
        """Get the next input."""
        import time

        flow_run_inputs = self.filter_for_inputs_sync()
        if flow_run_inputs:
            return self.to_instance(flow_run_inputs[0])

        start_time = time.monotonic()
        while True:
            time.sleep(self.poll_interval)
            flow_run_inputs = self.filter_for_inputs_sync()
            if flow_run_inputs:
                return self.to_instance(flow_run_inputs[0])
            if (
                self.timeout is not None
                and time.monotonic() - start_time >= self.timeout
            ):
                raise TimeoutError("Timed out waiting for input")

    def to_instance(
        self, flow_run_input: "FlowRunInput"
    ) -> Union[T, AutomaticRunInput[T]]:
        run_input = self.run_input_cls.load_from_flow_run_input(flow_run_input)

        if self.with_metadata:
            return run_input
        return run_input.value


async def _asend_input(
    flow_run_id: UUID,
    run_input: RunInput | pydantic.BaseModel,
    sender: Optional[str] = None,
    key_prefix: Optional[str] = None,
) -> None:
    """Send input to a flow run asynchronously (internal helper)."""
    _run_input: Union[RunInput, AutomaticRunInput[Any]]
    if isinstance(run_input, RunInput):
        _run_input = run_input
    else:
        input_cls: Type[AutomaticRunInput[Any]] = run_input_subclass_from_type(
            type(run_input)
        )
        _run_input = input_cls(value=run_input)

    if key_prefix is None:
        key_prefix = f"{_run_input.__class__.__name__.lower()}-auto"

    key = f"{key_prefix}-{uuid4()}"

    await acreate_flow_run_input_from_model(
        key=key, flow_run_id=flow_run_id, model_instance=_run_input, sender=sender
    )


def _send_input_sync(
    flow_run_id: UUID,
    run_input: RunInput | pydantic.BaseModel,
    sender: Optional[str] = None,
    key_prefix: Optional[str] = None,
) -> None:
    """Send input to a flow run synchronously (internal helper)."""
    _run_input: Union[RunInput, AutomaticRunInput[Any]]
    if isinstance(run_input, RunInput):
        _run_input = run_input
    else:
        input_cls: Type[AutomaticRunInput[Any]] = run_input_subclass_from_type(
            type(run_input)
        )
        _run_input = input_cls(value=run_input)

    if key_prefix is None:
        key_prefix = f"{_run_input.__class__.__name__.lower()}-auto"

    key = f"{key_prefix}-{uuid4()}"

    create_flow_run_input_from_model(
        key=key, flow_run_id=flow_run_id, model_instance=_run_input, sender=sender
    )


async def asend_input(
    run_input: Any,
    flow_run_id: UUID,
    sender: Optional[str] = None,
    key_prefix: Optional[str] = None,
) -> None:
    """Send input to a flow run asynchronously."""
    await _asend_input(
        flow_run_id=flow_run_id,
        run_input=run_input,
        sender=sender,
        key_prefix=key_prefix,
    )


@async_dispatch(asend_input)
def send_input(
    run_input: Any,
    flow_run_id: UUID,
    sender: Optional[str] = None,
    key_prefix: Optional[str] = None,
) -> None:
    """Send input to a flow run."""
    _send_input_sync(
        flow_run_id=flow_run_id,
        run_input=run_input,
        sender=sender,
        key_prefix=key_prefix,
    )


@overload
def receive_input(  # type: ignore[overload-overlap]
    input_type: Union[Type[R], pydantic.BaseModel],
    timeout: Optional[float] = 3600,
    poll_interval: float = 10,
    raise_timeout_error: bool = False,
    exclude_keys: Optional[Set[str]] = None,
    key_prefix: Optional[str] = None,
    flow_run_id: Optional[UUID] = None,
    with_metadata: bool = False,
) -> GetInputHandler[R]: ...


@overload
def receive_input(
    input_type: Type[T],
    timeout: Optional[float] = 3600,
    poll_interval: float = 10,
    raise_timeout_error: bool = False,
    exclude_keys: Optional[Set[str]] = None,
    key_prefix: Optional[str] = None,
    flow_run_id: Optional[UUID] = None,
    with_metadata: bool = False,
) -> GetAutomaticInputHandler[T]: ...


def receive_input(
    input_type: Union[Type[R], Type[T], pydantic.BaseModel],
    timeout: Optional[float] = 3600,
    poll_interval: float = 10,
    raise_timeout_error: bool = False,
    exclude_keys: Optional[Set[str]] = None,
    key_prefix: Optional[str] = None,
    flow_run_id: Optional[UUID] = None,
    with_metadata: bool = False,
) -> Union[GetAutomaticInputHandler[T], GetInputHandler[R]]:
    # The typing in this module is a bit complex, and at this point `mypy`
    # thinks that `run_input_subclass_from_type` accepts a `Type[Never]` but
    # the signature is the same as here:
    #   Union[Type[R], Type[T], pydantic.BaseModel],
    # Seems like a possible mypy bug, so we'll ignore the type check here.
    input_cls: Union[Type[AutomaticRunInput[T]], Type[R]] = (
        run_input_subclass_from_type(input_type)
    )  # type: ignore[arg-type]

    if issubclass(input_cls, AutomaticRunInput):
        return input_cls.receive(
            timeout=timeout,
            poll_interval=poll_interval,
            raise_timeout_error=raise_timeout_error,
            exclude_keys=exclude_keys,
            key_prefix=key_prefix,
            flow_run_id=flow_run_id,
            with_metadata=with_metadata,
        )
    else:
        return input_cls.receive(
            timeout=timeout,
            poll_interval=poll_interval,
            raise_timeout_error=raise_timeout_error,
            exclude_keys=exclude_keys,
            key_prefix=key_prefix,
            flow_run_id=flow_run_id,
        )



from __future__ import annotations

from functools import partial
from typing import Annotated, ClassVar, Union

from pydantic import AliasChoices, AliasPath, BeforeValidator, Field
from pydantic_settings import SettingsConfigDict

from prefect.settings.base import PrefectBaseSettings, build_settings_config
from prefect.types import validate_set_T_from_delim_string


class PluginsSettings(PrefectBaseSettings):
    """
    Settings for configuring the experimental plugin system
    """

    model_config: ClassVar[SettingsConfigDict] = build_settings_config(
        ("experiments", "plugins")
    )

    enabled: bool = Field(
        default=False,
        description="Enable the experimental plugin system.",
    )

    allow: Annotated[
        Union[set[str], None],
        BeforeValidator(partial(validate_set_T_from_delim_string, type_=str)),
    ] = Field(
        default=None,
        description="Comma-separated list of plugin names to allow. If set, only these plugins will be loaded.",
    )

    deny: Annotated[
        Union[set[str], None],
        BeforeValidator(partial(validate_set_T_from_delim_string, type_=str)),
    ] = Field(
        default=None,
        description="Comma-separated list of plugin names to deny. These plugins will not be loaded.",
    )

    setup_timeout_seconds: float = Field(
        default=20.0,
        description="Maximum time in seconds for all plugins to complete their setup hooks.",
    )

    strict: bool = Field(
        default=False,
        description="If True, exit if a required plugin fails during setup.",
    )

    safe_mode: bool = Field(
        default=False,
        description="If True, load plugins but do not execute their hooks. Useful for testing.",
    )


class ExperimentsSettings(PrefectBaseSettings):
    """
    Settings for configuring experimental features
    """

    model_config: ClassVar[SettingsConfigDict] = build_settings_config(("experiments",))

    warn: bool = Field(
        default=True,
        description="If `True`, warn on usage of experimental features.",
        validation_alias=AliasChoices(
            AliasPath("warn"), "prefect_experiments_warn", "prefect_experimental_warn"
        ),
    )

    plugins: PluginsSettings = Field(
        default_factory=PluginsSettings,
        description="Settings for the experimental plugin system",
    )



from __future__ import annotations

from datetime import datetime, timedelta, timezone
from uuid import UUID

from prefect.server.concurrency.lease_storage import (
    ConcurrencyLeaseHolder,
    ConcurrencyLimitLeaseMetadata,
)
from prefect.server.concurrency.lease_storage import (
    ConcurrencyLeaseStorage as _ConcurrencyLeaseStorage,
)
from prefect.server.utilities.leasing import ResourceLease


class ConcurrencyLeaseStorage(_ConcurrencyLeaseStorage):
    """
    A singleton concurrency lease storage implementation that stores leases in memory.
    """

    _instance: "ConcurrencyLeaseStorage | None" = None
    _initialized: bool = False

    def __new__(cls) -> "ConcurrencyLeaseStorage":
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self):
        if self.__class__._initialized:
            return

        self.leases: dict[UUID, ResourceLease[ConcurrencyLimitLeaseMetadata]] = {}
        self.expirations: dict[UUID, datetime] = {}
        self.__class__._initialized = True

    async def create_lease(
        self,
        resource_ids: list[UUID],
        ttl: timedelta,
        metadata: ConcurrencyLimitLeaseMetadata | None = None,
    ) -> ResourceLease[ConcurrencyLimitLeaseMetadata]:
        expiration = datetime.now(timezone.utc) + ttl
        lease = ResourceLease(
            resource_ids=resource_ids, metadata=metadata, expiration=expiration
        )
        self.leases[lease.id] = lease
        self.expirations[lease.id] = expiration
        return lease

    async def read_lease(
        self, lease_id: UUID
    ) -> ResourceLease[ConcurrencyLimitLeaseMetadata] | None:
        return self.leases.get(lease_id)

    async def renew_lease(self, lease_id: UUID, ttl: timedelta) -> bool:
        """
        Atomically renew a concurrency lease by updating its expiration.

        Checks if the lease exists before updating the expiration index,
        preventing orphaned index entries.

        Args:
            lease_id: The ID of the lease to renew
            ttl: The new time-to-live duration

        Returns:
            True if the lease was renewed, False if it didn't exist
        """
        if lease_id not in self.leases:
            # Clean up any orphaned expiration entry
            self.expirations.pop(lease_id, None)
            return False

        self.expirations[lease_id] = datetime.now(timezone.utc) + ttl
        return True

    async def revoke_lease(self, lease_id: UUID) -> None:
        self.leases.pop(lease_id, None)
        self.expirations.pop(lease_id, None)

    async def read_active_lease_ids(
        self, limit: int = 100, offset: int = 0
    ) -> list[UUID]:
        now = datetime.now(timezone.utc)
        active_leases = [
            lease_id
            for lease_id, expiration in self.expirations.items()
            if expiration > now
        ]
        return active_leases[offset : offset + limit]

    async def read_expired_lease_ids(self, limit: int = 100) -> list[UUID]:
        now = datetime.now(timezone.utc)
        expired_leases = [
            lease_id
            for lease_id, expiration in self.expirations.items()
            if expiration < now
        ]
        return expired_leases[:limit]

    async def list_holders_for_limit(
        self, limit_id: UUID
    ) -> list[tuple[UUID, ConcurrencyLeaseHolder]]:
        """List all holders for a given concurrency limit."""
        now = datetime.now(timezone.utc)
        holders_with_leases: list[tuple[UUID, ConcurrencyLeaseHolder]] = []

        for lease_id, lease in self.leases.items():
            # Check if lease is active and for the specified limit
            if (
                limit_id in lease.resource_ids
                and self.expirations.get(lease_id, now) > now
                and lease.metadata
                and lease.metadata.holder
            ):
                holders_with_leases.append((lease.id, lease.metadata.holder))

        return holders_with_leases



"""
Cloud command — authenticate and interact with Prefect Cloud.
"""

from __future__ import annotations

import os
from typing import TYPE_CHECKING, Annotated

import cyclopts

import prefect.cli._app as _cli
from prefect.cli._cloud_utils import (
    check_key_is_valid_for_login,
    confirm_logged_in,
    get_current_workspace,
    login_with_browser,
    prompt_for_account_and_workspace,
    prompt_select_from_list,
)
from prefect.cli._utilities import (
    exit_with_error,
    exit_with_success,
    with_cli_exception_handling,
)

if TYPE_CHECKING:
    from prefect.client.schemas import Workspace

cloud_app: cyclopts.App = cyclopts.App(
    name="cloud",
    help="Authenticate and interact with Prefect Cloud.",
    version_flags=[],
    help_flags=["--help"],
)

# --- workspace sub-app ---
workspace_app: cyclopts.App = cyclopts.App(
    name="workspace",
    alias="workspaces",
    help="View and set Prefect Cloud Workspaces.",
    version_flags=[],
    help_flags=["--help"],
)
cloud_app.command(workspace_app)


# =====================================================================
# cloud login / logout
# =====================================================================


@cloud_app.command(name="login")
@with_cli_exception_handling
async def login(
    *,
    key: Annotated[
        str | None,
        cyclopts.Parameter(
            "--key", alias="-k", help="API Key to authenticate with Prefect"
        ),
    ] = None,
    workspace_handle: Annotated[
        str | None,
        cyclopts.Parameter(
            "--workspace",
            alias="-w",
            help="Full handle of workspace, in format '<account_handle>/<workspace_handle>'",
        ),
    ] = None,
):
    """Log in to Prefect Cloud.
    Creates a new profile configured to use the specified PREFECT_API_KEY.
    Uses a previously configured profile if it exists.
    """
    import httpx

    from prefect.cli._prompts import confirm
    from prefect.client.cloud import CloudUnauthorizedError, get_cloud_client
    from prefect.context import get_settings_context
    from prefect.settings import (
        PREFECT_API_KEY,
        PREFECT_API_URL,
        PREFECT_CLOUD_UI_URL,
        load_profiles,
        save_profiles,
        update_current_profile,
    )
    from prefect.utilities.collections import listrepr

    console = _cli.console

    if not _cli.is_interactive() and (not key or not workspace_handle):
        exit_with_error(
            "When not using an interactive terminal, you must supply a `--key` and"
            " `--workspace`."
        )

    profiles = load_profiles()
    current_profile = get_settings_context().profile
    env_var_api_key = os.getenv("PREFECT_API_KEY")
    selected_workspace: Workspace | None = None

    if env_var_api_key and key and env_var_api_key != key:
        exit_with_error(
            "Cannot log in with a key when a different PREFECT_API_KEY is present as an"
            " environment variable that will override it."
        )

    if key and env_var_api_key and env_var_api_key == key:
        is_valid_key = await check_key_is_valid_for_login(key)
        is_correct_key_format = key.startswith("pnu_") or key.startswith("pnb_")
        if not is_valid_key:
            help_message = "Please ensure your credentials are correct and unexpired."
            if not is_correct_key_format:
                help_message = "Your key is not in our expected format."
            exit_with_error(
                f"Unable to authenticate with Prefect Cloud. {help_message}"
            )

    already_logged_in_profiles: list[str] = []
    for name, profile in profiles.items():
        profile_key = profile.settings.get(PREFECT_API_KEY)
        if (
            # If a key is provided, only show profiles with the same key
            (key and profile_key == key)
            # Otherwise, show all profiles with a key set
            or (not key and profile_key is not None)
            # Check that the key is usable to avoid suggesting unauthenticated profiles
            and await check_key_is_valid_for_login(profile_key)
        ):
            already_logged_in_profiles.append(name)

    current_profile_is_logged_in = current_profile.name in already_logged_in_profiles

    if current_profile_is_logged_in:
        console.print("It looks like you're already authenticated on this profile.")
        if _cli.is_interactive():
            should_reauth = confirm(
                "Would you like to reauthenticate?", default=False, console=console
            )
        else:
            should_reauth = True

        if not should_reauth:
            console.print("Using the existing authentication on this profile.")
            key = PREFECT_API_KEY.value()

    elif already_logged_in_profiles:
        console.print(
            "It looks like you're already authenticated with another profile."
        )
        if confirm(
            "Would you like to switch profiles?",
            default=True,
            console=console,
        ):
            profile_name = prompt_select_from_list(
                console,
                "Which authenticated profile would you like to switch to?",
                already_logged_in_profiles,
            )

            profiles.set_active(profile_name)
            save_profiles(profiles)
            exit_with_success(f"Switched to authenticated profile {profile_name!r}.")

    if not key:
        choice = prompt_select_from_list(
            console,
            "How would you like to authenticate?",
            [
                ("browser", "Log in with a web browser"),
                ("key", "Paste an API key"),
            ],
        )

        if choice == "key":
            from prefect.cli._prompts import prompt as _prompt

            key = _prompt("Paste your API key", password=True, console=console)
        elif choice == "browser":
            key = await login_with_browser(console)

    if TYPE_CHECKING:
        assert isinstance(key, str)

    async with get_cloud_client(api_key=key) as client:
        try:
            workspaces = await client.read_workspaces()
            current_workspace = get_current_workspace(workspaces)
            prompt_switch_workspace = False
        except CloudUnauthorizedError:
            if key.startswith("pcu"):
                help_message = (
                    "It looks like you're using API key from Cloud 1"
                    " (https://cloud.prefect.io). Make sure that you generate API key"
                    " using Cloud 2 (https://app.prefect.cloud)"
                )
            elif not key.startswith("pnu_") and not key.startswith("pnb_"):
                help_message = (
                    "Your key is not in our expected format: 'pnu_' or 'pnb_'."
                )
            else:
                help_message = (
                    "Please ensure your credentials are correct and unexpired."
                )
            exit_with_error(
                f"Unable to authenticate with Prefect Cloud. {help_message}"
            )
        except httpx.HTTPStatusError as exc:
            exit_with_error(f"Error connecting to Prefect Cloud: {exc!r}")

    if workspace_handle:
        # Search for the given workspace
        for workspace in workspaces:
            if workspace.handle == workspace_handle:
                selected_workspace = workspace
                break
        else:
            if workspaces:
                hint = (
                    " Available workspaces:"
                    f" {listrepr((w.handle for w in workspaces), ', ')}"
                )
            else:
                hint = ""

            exit_with_error(f"Workspace {workspace_handle!r} not found." + hint)
    else:
        # Prompt a switch if the number of workspaces is greater than one
        prompt_switch_workspace = len(workspaces) > 1

        # Confirm that we want to switch if the current profile is already logged in
        if (
            current_profile_is_logged_in and current_workspace is not None
        ) and prompt_switch_workspace:
            console.print(
                f"You are currently using workspace {current_workspace.handle!r}."
            )
            prompt_switch_workspace = confirm(
                "Would you like to switch workspaces?", default=False, console=console
            )
    if prompt_switch_workspace:
        go_back = True
        while go_back:
            selected_workspace, go_back = await prompt_for_account_and_workspace(
                workspaces, console
            )
        if selected_workspace is None:
            exit_with_error("No workspace selected.")

    elif not selected_workspace and not workspace_handle:
        if current_workspace:
            selected_workspace = current_workspace
        elif len(workspaces) > 0:
            selected_workspace = workspaces[0]
        else:
            exit_with_error(
                "No workspaces found! Create a workspace at"
                f" {PREFECT_CLOUD_UI_URL.value()} and try again."
            )

    if TYPE_CHECKING:
        from prefect.client.schemas import Workspace as WorkspaceType

        assert isinstance(selected_workspace, WorkspaceType)

    update_current_profile(
        {
            PREFECT_API_KEY: key,
            PREFECT_API_URL: selected_workspace.api_url(),
        }
    )

    exit_with_success(
        f"Authenticated with Prefect Cloud! Using workspace {selected_workspace.handle!r}."
    )


@cloud_app.command(name="logout")
@with_cli_exception_handling
async def logout():
    """Logout the current workspace.
    Reset PREFECT_API_KEY and PREFECT_API_URL to default.
    """
    from prefect.context import get_settings_context
    from prefect.settings import (
        PREFECT_API_KEY,
        PREFECT_API_URL,
        update_current_profile,
    )

    current_profile = get_settings_context().profile

    if current_profile.settings.get(PREFECT_API_KEY) is None:
        exit_with_error("Current profile is not logged into Prefect Cloud.")

    update_current_profile(
        {
            PREFECT_API_URL: None,
            PREFECT_API_KEY: None,
        },
    )

    exit_with_success("Logged out from Prefect Cloud.")


# =====================================================================
# cloud workspace
# =====================================================================


@workspace_app.command(name="ls")
@with_cli_exception_handling
async def workspace_ls():
    """List available workspaces."""
    from rich.table import Table

    from prefect.client.cloud import CloudUnauthorizedError, get_cloud_client

    confirm_logged_in(console=_cli.console)

    async with get_cloud_client() as client:
        try:
            workspaces = await client.read_workspaces()
        except CloudUnauthorizedError:
            exit_with_error(
                "Unable to authenticate. Please ensure your credentials are correct."
            )

    current_workspace = get_current_workspace(workspaces)

    table = Table(caption="* active workspace")
    table.add_column(
        "[#024dfd]Workspaces:", justify="left", style="#8ea0ae", no_wrap=True
    )

    for workspace_handle in sorted(workspace.handle for workspace in workspaces):
        if current_workspace and workspace_handle == current_workspace.handle:
            table.add_row(f"[green]* {workspace_handle}[/green]")
        else:
            table.add_row(f"  {workspace_handle}")

    _cli.console.print(table)


@workspace_app.command(name="set")
@with_cli_exception_handling
async def workspace_set(
    *,
    workspace_handle: Annotated[
        str | None,
        cyclopts.Parameter(
            "--workspace",
            alias="-w",
            help="Full handle of workspace, in format '<account_handle>/<workspace_handle>'",
        ),
    ] = None,
):
    """Set current workspace. Shows a workspace picker if no workspace is specified."""
    from prefect.client.cloud import CloudUnauthorizedError, get_cloud_client
    from prefect.settings import PREFECT_API_URL, update_current_profile

    confirm_logged_in(console=_cli.console)
    console = _cli.console

    async with get_cloud_client() as client:
        try:
            workspaces = await client.read_workspaces()
        except CloudUnauthorizedError:
            exit_with_error(
                "Unable to authenticate. Please ensure your credentials are correct."
            )

        if workspace_handle:
            # Search for the given workspace
            for workspace in workspaces:
                if workspace.handle == workspace_handle:
                    break
            else:
                exit_with_error(f"Workspace {workspace_handle!r} not found.")
        else:
            if not workspaces:
                exit_with_error("No workspaces found in the selected account.")

            # Store the original list of workspaces
            original_workspaces = workspaces.copy()

            go_back = True
            workspace = None
            loop_count = 0

            while go_back:
                loop_count += 1

                # If we're going back, use the original list of workspaces
                if loop_count > 1:
                    workspaces = original_workspaces.copy()

                workspace, go_back = await prompt_for_account_and_workspace(
                    workspaces, console
                )

            if workspace is None:
                exit_with_error("No workspace selected.")

        profile = update_current_profile({PREFECT_API_URL: workspace.api_url()})
        exit_with_success(
            f"Successfully set workspace to {workspace.handle!r} in profile"
            f" {profile.name!r}."
        )


# =====================================================================
# Sub-app registrations
# =====================================================================

from prefect.cli.cloud.asset import asset_app  # noqa: E402
from prefect.cli.cloud.ip_allowlist import ip_allowlist_app  # noqa: E402
from prefect.cli.cloud.webhook import webhook_app  # noqa: E402

cloud_app.command(webhook_app)
cloud_app.command(ip_allowlist_app)
cloud_app.command(asset_app)



"""
Flow command — native cyclopts implementation.

View and serve flows.
"""

from typing import Annotated, Optional

import cyclopts
import orjson
from rich.table import Table

import prefect.cli._app as _cli
from prefect.cli._utilities import (
    exit_with_error,
    with_cli_exception_handling,
)
from prefect.client.orchestration import get_client
from prefect.client.schemas.actions import DeploymentScheduleCreate
from prefect.client.schemas.schedules import construct_schedule
from prefect.client.schemas.sorting import FlowSort
from prefect.deployments.runner import RunnerDeployment
from prefect.exceptions import MissingFlowError
from prefect.runner import Runner
from prefect.utilities import urls

flow_app: cyclopts.App = cyclopts.App(
    name="flow",
    alias="flows",
    help="View and serve flows.",
    version_flags=[],
    help_flags=["--help"],
)


@flow_app.command()
@with_cli_exception_handling
async def ls(
    *,
    limit: Annotated[
        int,
        cyclopts.Parameter("--limit", help="Maximum number of flows to list."),
    ] = 15,
    output: Annotated[
        Optional[str],
        cyclopts.Parameter(
            "--output",
            alias="-o",
            help="Specify an output format. Currently supports: json",
        ),
    ] = None,
):
    """View flows."""
    if output and output.lower() != "json":
        exit_with_error("Only 'json' output format is supported.")

    async with get_client() as client:
        flows = await client.read_flows(
            limit=limit,
            sort=FlowSort.CREATED_DESC,
        )

    if not flows:
        if output and output.lower() == "json":
            _cli.console.print("[]")
            return
        _cli.console.print("No flows found.")
        return

    if output and output.lower() == "json":
        flows_json = [f.model_dump(mode="json") for f in flows]
        json_output = orjson.dumps(flows_json, option=orjson.OPT_INDENT_2).decode()
        _cli.console.print(json_output, soft_wrap=True)
    else:
        table = Table(title="Flows")
        table.add_column("ID", justify="right", style="cyan", no_wrap=True)
        table.add_column("Name", style="green", no_wrap=True)
        table.add_column("Created", no_wrap=True)

        for flow in flows:
            table.add_row(
                str(flow.id),
                str(flow.name),
                str(flow.created),
            )

        _cli.console.print(table)


@flow_app.command()
@with_cli_exception_handling
async def serve(
    entrypoint: str,
    *,
    name: Annotated[
        str,
        cyclopts.Parameter(
            "--name",
            alias="-n",
            help="The name to give the deployment created for the flow.",
        ),
    ],
    description: Annotated[
        Optional[str],
        cyclopts.Parameter(
            "--description",
            alias="-d",
            help=(
                "The description to give the created deployment. If not provided, the"
                " description will be populated from the flow's description."
            ),
        ),
    ] = None,
    version: Annotated[
        Optional[str],
        cyclopts.Parameter(
            "--version", alias="-v", help="A version to give the created deployment."
        ),
    ] = None,
    tag: Annotated[
        Optional[list[str]],
        cyclopts.Parameter(
            "--tag",
            alias="-t",
            help="One or more optional tags to apply to the created deployment.",
        ),
    ] = None,
    cron: Annotated[
        Optional[str],
        cyclopts.Parameter(
            "--cron",
            help=(
                "A cron string that will be used to set a schedule for the created"
                " deployment."
            ),
        ),
    ] = None,
    interval: Annotated[
        Optional[int],
        cyclopts.Parameter(
            "--interval",
            help=(
                "An integer specifying an interval (in seconds) between scheduled runs"
                " of the flow."
            ),
        ),
    ] = None,
    anchor_date: Annotated[
        Optional[str],
        cyclopts.Parameter(
            "--anchor-date", help="The start date for an interval schedule."
        ),
    ] = None,
    rrule: Annotated[
        Optional[str],
        cyclopts.Parameter(
            "--rrule",
            help="An RRule that will be used to set a schedule for the created deployment.",
        ),
    ] = None,
    timezone: Annotated[
        Optional[str],
        cyclopts.Parameter(
            "--timezone",
            help="Timezone to used scheduling flow runs e.g. 'America/New_York'",
        ),
    ] = None,
    pause_on_shutdown: Annotated[
        bool,
        cyclopts.Parameter(
            "--pause-on-shutdown",
            help=(
                "If set, provided schedule will be paused when the serve command is"
                " stopped. If not set, the schedules will continue running."
            ),
        ),
    ] = True,
    limit: Annotated[
        Optional[int],
        cyclopts.Parameter(
            "--limit",
            help=(
                "The maximum number of runs that can be executed concurrently by the"
                " created runner; only applies to this served flow."
                " To apply a limit across multiple served flows, use global_limit."
            ),
        ),
    ] = None,
    global_limit: Annotated[
        Optional[int],
        cyclopts.Parameter(
            "--global-limit",
            help=(
                "The maximum number of concurrent runs allowed across all served"
                " flow instances associated with the same deployment."
            ),
        ),
    ] = None,
):
    """Serve a flow via an entrypoint."""
    runner = Runner(
        name=name,
        pause_on_shutdown=pause_on_shutdown,
        limit=limit,
    )
    try:
        schedules = []
        if interval or cron or rrule:
            schedule = construct_schedule(
                interval=interval,
                cron=cron,
                rrule=rrule,
                timezone=timezone,
                anchor_date=anchor_date,
            )
            schedules = [DeploymentScheduleCreate(schedule=schedule, active=True)]

        runner_deployment = RunnerDeployment.from_entrypoint(
            entrypoint=entrypoint,
            name=name,
            schedules=schedules,
            description=description,
            tags=tag or [],
            version=version,
            concurrency_limit=global_limit,
        )
    except (MissingFlowError, ValueError) as exc:
        exit_with_error(str(exc))

    deployment_id = await runner.add_deployment(runner_deployment)

    help_message = (
        f"[green]Your flow {runner_deployment.flow_name!r} is being served and polling"
        " for scheduled runs!\n[/]\nTo trigger a run for this flow, use the following"
        " command:\n[blue]\n\t$ prefect deployment run"
        f" '{runner_deployment.flow_name}/{name}'\n[/]"
    )

    deployment_url = urls.url_for("deployment", obj_id=deployment_id)
    if deployment_url:
        help_message += (
            "\nYou can also run your flow via the Prefect UI:"
            f" [blue]{deployment_url}[/]\n"
        )

    _cli.console.print(help_message, soft_wrap=True)
    await runner.start()



"""Shared worker utilities used by both typer and cyclopts CLI implementations."""

from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Type

from prefect._internal.integrations import KNOWN_EXTRAS_FOR_PACKAGES
from prefect.client.collections import get_collections_metadata_client
from prefect.client.orchestration import get_client
from prefect.client.schemas.filters import WorkQueueFilter, WorkQueueFilterName
from prefect.exceptions import ObjectNotFound
from prefect.plugins import load_prefect_collections
from prefect.utilities.dispatch import lookup_type
from prefect.workers.base import BaseWorker

if TYPE_CHECKING:
    from rich.console import Console


async def _check_work_pool_paused(work_pool_name: str) -> bool:
    try:
        async with get_client() as client:
            work_pool = await client.read_work_pool(work_pool_name=work_pool_name)
            return work_pool.is_paused
    except ObjectNotFound:
        return False


async def _check_work_queues_paused(
    work_pool_name: str, work_queues: Optional[list[str]]
) -> bool:
    """Check if all work queues in the work pool are paused.

    Args:
        work_pool_name: the name of the work pool to check
        work_queues: the names of the work queues to check
    """
    try:
        work_queues_filter = (
            WorkQueueFilter(name=WorkQueueFilterName(any_=work_queues))
            if work_queues
            else None
        )
        async with get_client() as client:
            wqs = await client.read_work_queues(
                work_pool_name=work_pool_name, work_queue_filter=work_queues_filter
            )
            return all(queue.is_paused for queue in wqs) if wqs else False
    except ObjectNotFound:
        return False


async def _retrieve_worker_type_from_pool(
    console: "Console",
    exit_fn: object,
    work_pool_name: Optional[str] = None,
) -> str:
    """Discover the worker type for a work pool.

    Args:
        console: Rich console for output
        exit_fn: Called with an error message when the pool is push/managed
        work_pool_name: Name of the work pool
    """
    try:
        async with get_client() as client:
            work_pool = await client.read_work_pool(work_pool_name=work_pool_name)

        worker_type = work_pool.type
        console.print(
            f"Discovered type {worker_type!r} for work pool {work_pool.name!r}."
        )

        if work_pool.is_push_pool or work_pool.is_managed_pool:
            exit_fn(
                "Workers are not required for push work pools. "
                "See https://docs.prefect.io/latest/deploy/infrastructure-examples/serverless "
                "for more details."
            )

    except ObjectNotFound:
        console.print(
            (
                f"Work pool {work_pool_name!r} does not exist and no worker type was"
                " provided. Starting a process worker..."
            ),
            style="yellow",
        )
        worker_type = "process"
    return worker_type


def _load_worker_class(worker_type: str) -> Optional[Type[BaseWorker]]:
    try:
        load_prefect_collections()
        return lookup_type(BaseWorker, worker_type)
    except KeyError:
        return None


async def _install_package(
    console: "Console",
    package: str,
    upgrade: bool = False,
) -> None:
    from prefect._internal.installation import ainstall_packages

    console.print(f"Installing {package}...")
    install_package = KNOWN_EXTRAS_FOR_PACKAGES.get(package, package)
    await ainstall_packages([install_package], stream_output=True, upgrade=upgrade)


async def _find_package_for_worker_type(
    console: "Console",
    worker_type: str,
) -> Optional[str]:
    async with get_collections_metadata_client() as client:
        worker_metadata = await client.read_worker_metadata()

    worker_types_with_packages = {
        wt: package_name
        for package_name, worker_dict in worker_metadata.items()
        for wt in worker_dict
        if wt != "prefect-agent"
    }
    try:
        return worker_types_with_packages[worker_type]
    except KeyError:
        console.print(
            f"Could not find a package for worker type {worker_type!r}.",
            style="yellow",
        )
        return None



from prefect.assets.core import Asset, AssetProperties, add_asset_metadata
from prefect.assets.materialize import materialize

__all__ = ["Asset", "AssetProperties", "materialize", "add_asset_metadata"]
